"""Module for "singleton" QUERIES dictionary"""

import asyncio


_QUEUES = {}


def initialize_service_queue(service_name):
    """Initialize an `asyncio.Queue()` for a service

    Repeated calls to this function return the same queue

    Parameters
    ----------
    service_name : str
        Name of service to initialize queue for.

    Returns
    -------
    asyncio.Queue()
        Queue instance for this service.
    """
    return _QUEUES.setdefault(service_name, asyncio.Queue())


def tear_down_service_queue(service_name):
    """Remove the queue for a service

    The queue does not have to exist, so repeated calls to this function
    are OK.

    Parameters
    ----------
    service_name : str
        Name of service to delete queue for.
    """
    _QUEUES.pop(service_name, None)


def get_service_queue(service_name):
    """Retrieve the queue for a service

    Parameters
    ----------
    service_name : str
        Name of service to retrieve queue for.

    Returns
    -------
    asyncio.Queue or None
        Queue instance for this service, or `None` if the queue was not
        initialized.
    """
    return _QUEUES.get(service_name)



"""COMPASS Ordinance Location Validation logic

These are primarily used to validate that a legal document applies to a
particular location.
"""

import asyncio
import logging
from abc import ABC, abstractmethod

from compass.extraction.ngrams import convert_text_to_sentence_ngrams
from compass.utilities.enums import LLMUsageCategory


logger = logging.getLogger(__name__)


class LocationValidator(ABC):
    """Validation base class using a static system prompt"""

    SYSTEM_MESSAGE = None
    """LLM system message describing validation task"""

    def __init__(self, structured_llm_caller):
        """

        Parameters
        ----------
        structured_llm_caller : `StructuredLLMCaller`
            StructuredLLMCaller instance. Used for structured validation
            queries.
        """
        self.slc = structured_llm_caller

    async def check(self, content, **fmt_kwargs):
        """Check if the content passes the validation

        The exact validation is outlined in the class `SYSTEM_MESSAGE`.

        Parameters
        ----------
        content : str
            Document content to validate.
        **fmt_kwargs
            Keyword arguments to be passed to `SYSTEM_MESSAGE.format()`.

        Returns
        -------
        bool
            ``True`` if the content passes the validation check,
            ``False`` otherwise.
        """
        if not content:
            return False
        sys_msg = self.SYSTEM_MESSAGE.format(**fmt_kwargs)
        out = await self.slc.call(
            sys_msg,
            content,
            usage_sub_label=LLMUsageCategory.DOCUMENT_LOCATION_VALIDATION,
        )
        return self._parse_output(out)

    @abstractmethod
    def _parse_output(self, props):
        """Parse LLM response and return boolean validation response"""
        raise NotImplementedError


class URLValidator(LocationValidator):
    """Validator that checks whether a URL matches a county"""

    SYSTEM_MESSAGE = (
        "You extract structured data from a URL. Return your "
        "answer in JSON format. Your JSON file must include exactly two keys. "
        "The first key is 'correct_county', which is a boolean that is set to "
        "`True` if the URL mentions {county} County in some way. DO NOT infer "
        "based on information in the URL about any US state, city, township, "
        "or otherwise. `False` if not sure. The second key is "
        "'correct_state', which is a boolean that is set to `True` if the URL "
        "mentions {state} State in some way. DO NOT infer based on "
        "information in the URL about any US county, city, township, or "
        "otherwise. `False` if not sure."
    )

    def _parse_output(self, props):  # noqa: PLR6301
        """Parse LLM response and return boolean validation response"""
        logger.debug("Parsing URL validation output:\n\t%s", props)
        check_vars = ("correct_county", "correct_state")
        return all(props.get(var) for var in check_vars)


class CountyJurisdictionValidator(LocationValidator):
    """Validator that checks whether text applies at the county level"""

    SYSTEM_MESSAGE = (
        "You extract structured data from legal text. Return "
        "your answer in JSON format. Your JSON file must include exactly "
        "three keys. The first key is 'x', which is a boolean that is set to "
        "`True` if the text excerpt explicitly states that the legal "
        "regulations outlined within the text are applicable to a "
        "jurisdiction scope other than {county} County "
        "(i.e. they apply to a subdivision like a township or a city, or "
        "they apply more broadly, like to a state or the full country). "
        "`False` if the regulations in the text apply at the {county} County "
        "level, if the regulations in the text apply to all unincorporated "
        "areas of {county} County, or if there is not enough information to "
        "determine the answer. The second key is 'y', which is a boolean "
        "that is set to `True` if the text excerpt explicitly mentions that "
        "the regulations within apply to more than one county. `False` if "
        "the regulations in the text excerpt apply to a single county only "
        "or if there is not enough information to determine the answer. The "
        "third key is 'explanation', which is a string that contains a short "
        "explanation if you chose `True` for any answers above."
    )
    META_SCORE_KEY = "Jurisdiction Validation Score"

    def _parse_output(self, props):  # noqa: PLR6301
        """Parse LLM response and return boolean validation result"""
        logger.debug(
            "Parsing county jurisdiction validation output:\n\t%s", props
        )
        check_vars = ("x", "y")
        return not any(props.get(var) for var in check_vars)


class CountyNameValidator(LocationValidator):
    """Validator that checks whether text applies to a given county"""

    SYSTEM_MESSAGE = (
        "You extract structured data from legal text. Return "
        "your answer in JSON format. Your JSON file must include exactly "
        "three keys. The first key is 'wrong_county', which is a boolean that "
        "is set to `True` if the legal text is not for {county} County. Do "
        "not infer based on any information about any US state, city, "
        "township, or otherwise. `False` if the text applies to {county} "
        "County or if there is not enough information to determine the "
        "answer. The second key is 'wrong_state', which is a boolean that is "
        "set to `True` if the legal text is not for a county in {state} "
        "State. Do not infer based on any information about any US county, "
        "city, township, or otherwise. `False` if the text applies to "
        "a county in {state} State or if there is not enough information to "
        "determine the answer. The third key is 'explanation', which is a "
        "string that contains a short explanation if you chose `True` for "
        "any answers above."
    )
    META_SCORE_KEY = "Jurisdiction Name Validation Score"

    def _parse_output(self, props):  # noqa: PLR6301
        """Parse LLM response and return boolean validation response"""
        logger.debug("Parsing county validation output:\n\t%s", props)
        check_vars = ("wrong_county", "wrong_state")
        return not any(props.get(var) for var in check_vars)


class CountyValidator:
    """COMPASS Ordinance County validator

    Combines the logic of several validators into a single class.

    Purpose:
        Determine whether a document pertains to a specific county.
    Responsibilities:
        1. Use a combination of heuristics and LLM queries to determine
           whether or not a document pertains to a particular county.
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries and delegates sub-validation to
        :class:`~compass.validation.location.CountyNameValidator`,
        :class:`~compass.validation.location.CountyJurisdictionValidator`,
        and :class:`~compass.validation.location.URLValidator`.
    """

    def __init__(
        self, structured_llm_caller, score_thresh=0.8, text_splitter=None
    ):
        """

        Parameters
        ----------
        structured_llm_caller : `StructuredLLMCaller`
            StructuredLLMCaller instance. Used for structured validation
            queries.
        score_thresh : float, optional
            Score threshold to exceed when voting on content from raw
            pages. By default, ``0.8``.
        text_splitter : langchain.text_splitter.TextSplitter, optional
            Optional text splitter instance to attach to doc (used for
            splitting out pages in an HTML document).
            By default, ``None``.
        """
        self.score_thresh = score_thresh
        self.cn_validator = CountyNameValidator(structured_llm_caller)
        self.cj_validator = CountyJurisdictionValidator(structured_llm_caller)
        self.url_validator = URLValidator(structured_llm_caller)
        self.text_splitter = text_splitter

    async def check(self, doc, county, state):
        """Check if the document belongs to the county

        Parameters
        ----------
        doc : :class:`elm.web.document.BaseDocument`
            Document instance. Should contain a "source" key in the
            ``attrs`` that contains a URL (used for the URL validation
            check). Raw content will be parsed for county name and
            correct jurisdiction.
        county : str
            County that document should belong to.
        state : str
            State corresponding to `county` input.

        Returns
        -------
        bool
            `True` if the doc contents pertain to the input county.
            `False` otherwise.
        """
        if hasattr(doc, "text_splitter") and self.text_splitter is not None:
            old_splitter = doc.text_splitter
            doc.text_splitter = self.text_splitter
            out = await self._check(doc, county, state)
            doc.text_splitter = old_splitter
            return out

        return await self._check(doc, county, state)

    async def _check(self, doc, county, state):
        """Check if the document belongs to the county"""
        if self.text_splitter is not None:
            doc.text_splitter = self.text_splitter

        source = doc.attrs.get("source")
        logger.debug(
            "Validating document from source: %s", source or "Unknown"
        )
        logger.debug("Checking for correct for jurisdiction...")
        jurisdiction_is_county = await _validator_check_for_doc(
            validator=self.cj_validator,
            doc=doc,
            score_thresh=self.score_thresh,
            county=county,
        )
        if not jurisdiction_is_county:
            return False

        logger.debug(
            "Checking URL (%s) for county name...", source or "Unknown"
        )
        url_is_county = await self.url_validator.check(
            source, county=county, state=state
        )
        if url_is_county:
            return True

        logger.debug(
            "Checking text for county name (heuristic; URL: %s)...",
            source or "Unknown",
        )
        correct_county_heuristic = _heuristic_check_for_county_and_state(
            doc, county, state
        )
        logger.debug(
            "Found county name in text (heuristic): %s",
            correct_county_heuristic,
        )
        if correct_county_heuristic:
            return True

        logger.debug(
            "Checking text for county name (LLM; URL: %s)...",
            source or "Unknown",
        )
        return await _validator_check_for_doc(
            validator=self.cn_validator,
            doc=doc,
            score_thresh=self.score_thresh,
            county=county,
            state=state,
        )


def _heuristic_check_for_county_and_state(doc, county, state):
    """Check if county and state names are in doc"""
    return any(
        any(
            (county.lower() in fg and state.lower() in fg)
            for fg in convert_text_to_sentence_ngrams(t.lower(), 5)
        )
        for t in doc.pages
    )


async def _validator_check_for_doc(validator, doc, score_thresh=0.8, **kwargs):
    """Apply a validator check to a doc's raw pages"""
    outer_task_name = asyncio.current_task().get_name()
    validation_checks = [
        asyncio.create_task(
            validator.check(text, **kwargs), name=outer_task_name
        )
        for text in doc.raw_pages
    ]
    out = await asyncio.gather(*validation_checks)
    score = _weighted_vote(out, doc)
    doc.attrs[validator.META_SCORE_KEY] = score
    logger.debug(
        "%s is %.2f for doc from source %s (Pass: %s; threshold: %.2f)",
        validator.META_SCORE_KEY,
        score,
        doc.attrs.get("source", "Unknown"),
        str(score > score_thresh),
        score_thresh,
    )
    return score > score_thresh


def _weighted_vote(out, doc):
    """Compute weighted average of responses based on text length"""
    if not doc.raw_pages:
        return 0
    weights = [len(text) for text in doc.raw_pages]
    total = sum(
        verdict * weight for verdict, weight in zip(out, weights, strict=False)
    )
    return total / sum(weights)



"""COMPASS Ordinance Threaded services"""

import json
import shutil
import asyncio
import hashlib
from pathlib import Path
from abc import abstractmethod
from tempfile import TemporaryDirectory
from datetime import datetime, timedelta
from concurrent.futures import ThreadPoolExecutor

from elm.web.utilities import write_url_doc_to_file

from compass import COMPASS_DEBUG_LEVEL
from compass.services.base import Service
from compass.utilities import (
    LLM_COST_REGISTRY,
    extract_ord_year_from_doc_attrs,
    num_ordinances_in_doc,
)


def _cache_file_with_hash(doc, file_content, out_dir, make_name_unique=False):
    """Cache file and compute its hash"""
    cache_fp = write_url_doc_to_file(
        doc=doc,
        file_content=file_content,
        out_dir=out_dir,
        make_name_unique=make_name_unique,
    )
    return cache_fp, _compute_sha256(cache_fp)


def _compute_sha256(file_path):
    """Compute sha256 checksum for file on disk"""
    m = hashlib.sha256()
    m.update(Path(file_path).read_bytes())
    return f"sha256:{m.hexdigest()}"


def _move_file(doc, out_dir):
    """Move a file from a temp directory to an output directory"""
    cached_fp = doc.attrs.get("cache_fn")
    if cached_fp is None:
        return None

    cached_fp = Path(cached_fp)
    date = datetime.now().strftime("%Y_%m_%d")
    out_fn = doc.attrs.get("location_name", cached_fp.stem)
    out_fn = out_fn.replace(",", "").replace(" ", "_")
    out_fn = f"{out_fn}_downloaded_{date}"
    if not out_fn.endswith(cached_fp.suffix):
        out_fn = f"{out_fn}{cached_fp.suffix}"

    out_fp = Path(out_dir) / out_fn
    shutil.move(cached_fp, out_fp)
    return out_fp


def _write_cleaned_file(doc, out_dir):
    """Write cleaned ordinance text to directory"""
    location_name = doc.attrs.get("location_name")
    if location_name is None:
        return None

    out_dir = Path(out_dir)
    if COMPASS_DEBUG_LEVEL > 0:
        _write_interim_cleaned_files(doc, out_dir, location_name)

    key_to_fp = {
        "cleaned_ordinance_text": f"{location_name} Ordinance Summary.txt",
        "districts_text": f"{location_name} Districts.txt",
    }
    out_paths = []
    for key, fn in key_to_fp.items():
        cleaned_text = doc.attrs.get(key)
        if cleaned_text is None:
            continue

        out_fp = out_dir / fn
        out_fp.write_text(cleaned_text, encoding="utf-8")
        out_paths.append(out_fp)

    return out_paths


def _write_interim_cleaned_files(doc, out_dir, location_name):
    """Write intermediate output texts to file; helpful for debugging"""
    key_to_fp = {
        "ordinance_text": f"{location_name} Ordinance Original text.txt",
        "wind_energy_systems_text": f"{location_name} Wind Ordinance text.txt",
        "solar_energy_systems_text": (
            f"{location_name} Solar Ordinance text.txt"
        ),
        "permitted_use_text": (
            f"{location_name} Permitted Use Original text.txt"
        ),
        "permitted_use_only_text": (
            f"{location_name} Permitted Use Only text.txt"
        ),
    }
    for key, fn in key_to_fp.items():
        text = doc.attrs.get(key)
        if text is None:
            continue

        (out_dir / fn).write_text(text, encoding="utf-8")


def _write_ord_db(doc, out_dir):
    """Write parsed ordinance database to directory"""
    ord_db = doc.attrs.get("scraped_values")
    location_name = doc.attrs.get("location_name")

    if ord_db is None or location_name is None:
        return None

    out_fp = Path(out_dir) / f"{location_name} Ordinances.csv"
    ord_db.to_csv(out_fp, index=False)
    return out_fp


_PROCESSING_FUNCTIONS = {
    "move": _move_file,
    "write_clean": _write_cleaned_file,
    "write_db": _write_ord_db,
}


class ThreadedService(Service):
    """Service that contains a ThreadPoolExecutor instance"""

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        **kwargs
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        self._tpe_kwargs = kwargs or {}
        self.pool = None

    def acquire_resources(self):
        """Open thread pool and temp directory"""
        self.pool = ThreadPoolExecutor(**self._tpe_kwargs)

    def release_resources(self):
        """Shutdown thread pool and cleanup temp directory"""
        self.pool.shutdown(wait=True, cancel_futures=True)


class TempFileCache(ThreadedService):
    """Service that locally caches files downloaded from the internet"""

    def __init__(self, td_kwargs=None, tpe_kwargs=None):
        """

        Parameters
        ----------
        td_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`tempfile.TemporaryDirectory`. By default, ``None``.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self._td_kwargs = td_kwargs or {}
        self._td = None

    @property
    def can_process(self):
        """bool: Always ``True`` (limiting is handled by asyncio)"""
        return True

    def acquire_resources(self):
        """Open thread pool and temp directory"""
        super().acquire_resources()
        self._td = TemporaryDirectory(**self._td_kwargs)

    def release_resources(self):
        """Shutdown thread pool and cleanup temp directory"""
        self._td.cleanup()
        super().release_resources()

    async def process(self, doc, file_content, make_name_unique=False):
        """Write URL doc to file asynchronously

        Parameters
        ----------
        doc : elm.web.document.Document
            Document containing meta information about the file. Must
            have a "source" key in the ``attrs`` dict containing the
            URL, which will be converted to a file name using
            :func:`compute_fn_from_url`.
        file_content : str or bytes
            File content, typically string text for HTML files and bytes
            for PDF file.
        make_name_unique : bool, optional
            Option to make file name unique by adding a UUID at the end
            of the file name. By default, ``False``.

        Returns
        -------
        Path
            Path to output file.
        """
        loop = asyncio.get_running_loop()
        cache_fp, checksum = await loop.run_in_executor(
            self.pool,
            _cache_file_with_hash,
            doc,
            file_content,
            self._td.name,
            make_name_unique,
        )
        doc.attrs["checksum"] = checksum
        return cache_fp


class StoreFileOnDisk(ThreadedService):
    """Abstract service that manages the storage of a file on disk

    Storage can occur due to creation or a move of a file.
    """

    def __init__(self, out_dir, tpe_kwargs=None):
        """

        Parameters
        ----------
        out_dir : path-like
            Path to output directory where file should be stored.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self.out_dir = out_dir

    @property
    def can_process(self):
        """bool: Always ``True`` (limiting is handled by asyncio)"""
        return True

    async def process(self, doc):
        """Store file in out directory

        Parameters
        ----------
        doc : elm.web.document.Document
            Document containing meta information about the file. Must
            have relevant processing keys in the ``attrs`` dict,
            otherwise the file may not be stored in the output
            directory.

        Returns
        -------
        Path or None
            Path to output file, or `None` if no file was stored.
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            self.pool, _PROCESSING_FUNCTIONS[self._PROCESS], doc, self.out_dir
        )

    @property
    @abstractmethod
    def _PROCESS(self):  # noqa: N802
        """str: Key in `_PROCESSING_FUNCTIONS` defining the doc func"""
        raise NotImplementedError


class FileMover(StoreFileOnDisk):
    """Service that moves files to an output directory"""

    _PROCESS = "move"


class CleanedFileWriter(StoreFileOnDisk):
    """Service that writes cleaned text to a file"""

    _PROCESS = "write_clean"


class OrdDBFileWriter(StoreFileOnDisk):
    """Service that writes cleaned text to a file"""

    _PROCESS = "write_db"


class UsageUpdater(ThreadedService):
    """Service that updates usage info from a tracker into a file"""

    def __init__(self, usage_fp, tpe_kwargs=None):
        """

        Parameters
        ----------
        usage_fp : path-like
            Path to JSON file where usage should be tracked.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self.usage_fp = usage_fp
        self._is_processing = False

    @property
    def can_process(self):
        """bool: ``True`` if file not currently being written to"""
        return not self._is_processing

    async def process(self, tracker):
        """Add usage from tracker to file

        Any existing usage info in the file will remain unchanged
        EXCEPT for anything under the label of the input `tracker`,
        all of which will be replaced with info from the tracker itself.

        Parameters
        ----------
        tracker : elm.ods.services.usage.UsageTracker
            A usage tracker instance that contains usage info to be
            added to output file.
        """
        self._is_processing = True
        loop = asyncio.get_running_loop()
        out = await loop.run_in_executor(
            self.pool, _dump_usage, self.usage_fp, tracker
        )
        self._is_processing = False
        return out


class JurisdictionUpdater(ThreadedService):
    """Service that updates jurisdiction info into a file"""

    def __init__(self, jurisdiction_fp, tpe_kwargs=None):
        """

        Parameters
        ----------
        jurisdiction_fp : path-like
            Path to JSON file where jurisdictions should be tracked.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self.jurisdiction_fp = jurisdiction_fp
        self._is_processing = False

    @property
    def can_process(self):
        """bool: ``True`` if file not currently being written to"""
        return not self._is_processing

    async def process(self, county, doc, seconds_elapsed, usage_tracker=None):
        """Add usage from tracker to file

        Any existing usage info in the file will remain unchanged
        EXCEPT for anything under the label of the input `tracker`,
        all of which will be replaced with info from the tracker itself.

        Parameters
        ----------
        county : compass.utilities.location.Location
            County to record.
        doc : elm.web.document.Document
            Document containing meta information about the jurisdiction.
            Must have relevant processing keys in the ``attrs`` dict,
            otherwise the jurisdiction may not be recorded properly.
            If ``None``, the jurisdiction is assumed not to have been
            found.
        seconds_elapsed : int or float
            Total number of seconds it took to look for (and possibly
            parse) this document.
        usage_tracker : compass.services.usage.UsageTracker, optional
            Optional tracker instance to monitor token usage during
            LLM calls. By default, ``None``.
        """
        self._is_processing = True
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            self.pool,
            _dump_jurisdiction_info,
            self.jurisdiction_fp,
            county,
            doc,
            seconds_elapsed,
            usage_tracker,
        )
        self._is_processing = False


def _dump_usage(fp, tracker):
    """Dump usage to an existing file"""
    if not Path(fp).exists():
        usage_info = {}
    else:
        with Path.open(fp, encoding="utf-8") as fh:
            usage_info = json.load(fh)

    if tracker is not None:
        tracker.add_to(usage_info)

        with Path.open(fp, "w", encoding="utf-8") as fh:
            json.dump(usage_info, fh, indent=4)

    return usage_info


def _dump_jurisdiction_info(fp, county, doc, seconds_elapsed, usage_tracker):
    """Dump jurisdiction info to an existing file"""
    if not Path(fp).exists():
        jurisdiction_info = {"jurisdictions": []}
    else:
        with Path.open(fp, encoding="utf-8") as fh:
            jurisdiction_info = json.load(fh)

    new_info = {
        "full_name": county.full_name,
        "county": county.name,
        "state": county.state,
        "subdivision": None,
        "jurisdiction_type": None,
        "FIPS": county.fips,
        "found": False,
        "total_time": seconds_elapsed,
        "total_time_string": str(timedelta(seconds=seconds_elapsed)),
        "cost": None,
        "documents": None,
    }
    if usage_tracker is not None:
        cost = _compute_jurisdiction_cost(usage_tracker)
        new_info["cost"] = cost or None

    if num_ordinances_in_doc(doc) > 0:
        new_info["found"] = True
        new_info["documents"] = [_compile_doc_info(doc)]

    jurisdiction_info["jurisdictions"].append(new_info)
    with Path.open(fp, "w", encoding="utf-8") as fh:
        json.dump(jurisdiction_info, fh, indent=4)


def _compile_doc_info(doc):
    """Put together meta information about a single document"""
    return {
        "source": doc.attrs.get("source"),
        "ord_year": extract_ord_year_from_doc_attrs(doc.attrs),
        "ord_filename": Path(doc.attrs.get("out_fp", "Unknown")).name,
        "num_pages": len(doc.pages),
        "checksum": doc.attrs.get("checksum"),
    }


def _compute_jurisdiction_cost(usage_tracker):
    """Compute total cost from total tracked usage"""

    total_cost = 0
    for model, total_usage in usage_tracker.totals.items():
        model_costs = LLM_COST_REGISTRY.get(model, {})
        total_cost += (
            total_usage.get("prompt_tokens", 0)
            / 1e6
            * model_costs.get("prompt", 0)
        )
        total_cost += (
            total_usage.get("response_tokens", 0)
            / 1e6
            * model_costs.get("response", 0)
        )

    return total_cost



"""COMPASS Ordinances parsing utilities."""

import json
import logging


logger = logging.getLogger(__name__)
_ORD_CHECK_COLS = ["value", "adder", "min_dist", "max_dist", "summary"]


def clean_backticks_from_llm_response(content):
    """Remove markdown-style backticks from an LLM response

    Parameters
    ----------
    content : str
        LLM response that may or may not contain markdown-style triple
        backticks.

    Returns
    -------
    str
        LLM response stripped of the markdown-style backticks
    """
    content = content.lstrip().rstrip()
    return content.removeprefix("```").lstrip("\n").removesuffix("```")


def llm_response_as_json(content):
    """LLM response to JSON

    Parameters
    ----------
    content : str
        LLM response that contains a string representation of
        a JSON file.

    Returns
    -------
    dict
        Response parsed into dictionary. This dictionary will be empty
        if the response cannot be parsed by JSON.
    """
    content = clean_backticks_from_llm_response(content)
    content = content.removeprefix("json").lstrip("\n")
    content = content.replace("True", "true").replace("False", "false")
    try:
        content = json.loads(content)
    except json.decoder.JSONDecodeError:
        logger.exception(
            "LLM returned improperly formatted JSON. "
            "This is likely due to the completion running out of tokens. "
            "Setting a higher token limit may fix this error. "
            "Also ensure you are requesting JSON output in your prompt. "
            "JSON returned:\n%s",
            content,
        )
        content = {}
    return content


def merge_overlapping_texts(text_chunks, n=300):
    """Merge chunks of text by removing any overlap.

    Parameters
    ----------
    text_chunks : iterable of str
        Iterable containing text chunks which may or may not contain
        consecutive overlapping portions.
    n : int, optional
        Number of characters to check at the beginning of each message
        for overlap with the previous message. Will always be reduced to
        be less than or equal to half of the length of the previous
        chunk. By default, ``300``.

    Returns
    -------
    str
        Merged text.
    """
    text_chunks = list(filter(None, text_chunks))
    if not text_chunks:
        return ""

    out_text = text_chunks[0]
    for next_text in text_chunks[1:]:
        half_chunk_len = len(out_text) // 2
        check_len = min(n, half_chunk_len)
        next_chunks_start_ind = out_text[half_chunk_len:].find(
            next_text[:check_len]
        )
        if next_chunks_start_ind == -1:
            out_text = f"{out_text}\n{next_text}"
            continue
        next_chunks_start_ind += half_chunk_len
        out_text = "".join([out_text[:next_chunks_start_ind], next_text])
    return out_text


def extract_ord_year_from_doc_attrs(doc_attrs):
    """Extract year corresponding to the ordinance from doc instance

    Parameters
    ----------
    doc_attrs : dict
        Document meta information about the jurisdiction.
        Must have a "date" key in the attrs that is a tuple
        corresponding to the (year, month, day) of the ordinance to
        extract year successfully. If this key is missing, this function
        returns ``None``.

    Returns
    -------
    int or None
        Parsed year for ordinance (int) or ``None`` if it wasn't found
        in the document's attrs.
    """
    year = doc_attrs.get("date", (None, None, None))[0]
    return year if year is not None and year > 0 else None


def num_ordinances_in_doc(doc):
    """Count number of ordinances found in document

    Parameters
    ----------
    doc : elm.web.document.Document
        Document potentially containing ordinances for a jurisdiction.
        If no ordinance values are found, this function returns ``0``.

    Returns
    -------
    int
        Number of unique ordinance values extracted from this document.
    """
    if doc is None:
        return 0

    if "ordinance_values" not in doc.attrs:
        return 0

    return num_ordinances_dataframe(doc.attrs["ordinance_values"])


def num_ordinances_dataframe(data):
    """Count number of ordinances found in DataFrame

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame potentially containing ordinances for a jurisdiction.
        If no ordinance values are found, this function returns ``0``.

    Returns
    -------
    int
        Number of unique ordinance values extracted from this DataFrame.
    """
    if data is None:
        return 0

    if data.empty:
        return 0

    check_cols = [col for col in _ORD_CHECK_COLS if col in data]
    if not check_cols:
        return 0

    found_features = (~data[check_cols].isna()).to_numpy().sum(axis=1)
    return (found_features > 0).sum()



"""COMPASS service provider classes"""

import asyncio
import logging
import contextlib

from compass.services.queues import (
    initialize_service_queue,
    get_service_queue,
    tear_down_service_queue,
)
from compass.exceptions import COMPASSValueError


logger = logging.getLogger(__name__)


class _RunningProvider:
    """A running provider for a single service"""

    def __init__(self, service, queue):
        """

        Parameters
        ----------
        service : :class:`compass.services.base.Service`
            An instance of a single async service to run.
        queue : :class:`asyncio.Queue`
            Queue object for the running service.
        """
        self.service = service
        self.queue = queue
        self.jobs = set()

    async def run(self):
        """Run the service."""
        while True:
            await self.submit_jobs()
            await self.collect_responses()

    async def submit_jobs(self):
        """Submit jobs from the queue to processing

        The service can limit the number of submissions at a time by
        implementing the ``can_process`` property.

        If the queue is non-empty, the function takes jobs from it
        iteratively and submits them until the ``can_process`` property
        of the service returns ``False``. A call to ``can_process`` is
        submitted between every job pulled from the queue, so enure that
        method is performant. If the queue is empty, this function does
        one of two things:

            1) If there are no jobs processing, it waits on the queue
               to get more jobs and submits them as they come in
               (assuming the service allows it)
            2) If there are jobs processing, this function returns
               without waiting on more jobs from the queue.

        """
        if not self.service.can_process or self._q_empty_but_still_processing:
            return

        while self.service.can_process and self._can_fit_jobs:
            fut, outer_task_name, args, kwargs = await self.queue.get()
            task = asyncio.create_task(
                self.service.process_using_futures(fut, *args, **kwargs),
                name=outer_task_name,
            )
            self.queue.task_done()
            self.jobs.add(task)
            await _allow_service_to_update()

        return

    @property
    def _q_empty_but_still_processing(self):
        """bool: Queue empty but jobs still running (don't await)"""
        return self.queue.empty() and self.jobs

    @property
    def _can_fit_jobs(self):
        """bool: Job tracker not full"""
        return len(self.jobs) < self.service.MAX_CONCURRENT_JOBS

    async def collect_responses(self):
        """Collect responses from the service.

        This call will block further submissions to the service until
        at least one job finishes.
        """
        if not self.jobs:
            return

        complete, __ = await asyncio.wait(
            self.jobs, return_when=asyncio.FIRST_COMPLETED
        )

        for job in complete:
            self.jobs.remove(job)


class RunningAsyncServices:
    """Async context manager for running services."""

    def __init__(self, services):
        """

        Parameters
        ----------
        services : iterable
            An iterable of async services to run during program
            execution.
        """
        self.services = services
        self.__providers = []
        self._validate_services()

    def _validate_services(self):
        """Validate input services."""
        if len(self.services) < 1:
            msg = "Must provide at least one service to run!"
            raise COMPASSValueError(msg)

    def _reset_providers(self):
        """Reset running providers"""
        for c in self.__providers:
            c.cancel()
        self.__providers = []

    async def __aenter__(self):
        for service in self.services:
            logger.debug("Initializing Service: %s", service.name)
            with contextlib.suppress(AttributeError):
                logger.debug(
                    "    ↪ model_name=%r, rate_limit=%d",
                    service.model_name,
                    service.rate_limit,
                )
            queue = initialize_service_queue(service.name)
            service.acquire_resources()
            task = asyncio.create_task(_RunningProvider(service, queue).run())
            self.__providers.append(task)

    async def __aexit__(self, exc_type, exc, tb):
        try:
            for service in self.services:
                await get_service_queue(service.name).join()
                service.release_resources()
        finally:
            self._reset_providers()
            for service in self.services:
                logger.debug("Tearing down Service: %s", service.name)
                tear_down_service_queue(service.name)

    @classmethod
    def run(cls, services, coroutine):
        """Run an async function that relies on services.

        You can treat this function like the ``asyncio.run`` function
        with an extra parameter::

            openai_service = OpenAIService(...)
            RunningAsyncServices.run(
                [openai_service], my_async_func(*args, **kwargs)
            )


        Parameters
        ----------
        services : iterable of :class:`compass.services.base.Service`
            An iterable (i.e. a list) of Services that are needed to run
            the asynchronous function.
        coroutine :  coroutine
            A coroutine that should be run with the services.

        Returns
        -------
        Any
            Returns the output of the coroutine.
        """
        return asyncio.run(cls._run_coroutine(services, coroutine))

    @classmethod
    async def _run_coroutine(cls, services, coroutine):
        """Run a coroutine under services."""
        async with cls(services):
            return await coroutine


async def _allow_service_to_update():
    """Switch contexts, allowing service to update if it can process"""
    await asyncio.sleep(0)



"""Ordinance county info"""

import logging
from warnings import warn
from pathlib import Path

import pandas as pd

from compass.exceptions import COMPASSValueError


logger = logging.getLogger(__name__)
_COUNTY_DATA_FP = Path(__file__).parent.parent / "data" / "conus_counties.csv"


def load_all_county_info():
    """Load DataFrame containing info for all counties

    Returns
    -------
    pd.DataFrame
        DataFrame containing county info like names, FIPS, websites,
        etc. for all counties.
    """
    county_info = pd.read_csv(_COUNTY_DATA_FP)
    county_info = _convert_to_title(county_info, "County")
    return _convert_to_title(county_info, "State")


def county_websites(county_info=None):
    """Load mapping of county name and state to website

    Parameters
    ----------
    county_info : pd.DataFrame, optional
        DataFrame containing county names and websites. If ``None``,
        this info is loaded using :func:`load_county_info`.
        By default, ``None``.

    Returns
    -------
    dict
        Dictionary where keys are tuples of (county, state) and keys are
        the relevant website URL. Note that county and state names are
        lowercase.
    """
    if county_info is None:
        county_info = load_all_county_info()

    return {
        (row["County"].casefold(), row["State"].casefold()): row["Website"]
        for __, row in county_info.iterrows()
    }


def load_counties_from_fp(county_fp):
    """Load county info base don counties in the input fp

    Parameters
    ----------
    county_fp : path-like
        Path to csv file containing "County" and "State" columns that
        define the counties for which info should be loaded.

    Returns
    -------
    pd.DataFrame
        DataFrame containing county info like names, FIPS, websites,
        etc. for all requested counties (that were found).
    """
    counties = pd.read_csv(county_fp)
    _validate_county_input(counties)

    counties = _convert_to_title(counties, "County")
    counties = _convert_to_title(counties, "State")

    all_county_info = load_all_county_info()
    counties = counties.merge(
        all_county_info, on=["County", "State"], how="left"
    )

    counties = _filter_not_found_counties(counties)
    return _format_county_df_for_output(counties)


def _validate_county_input(df):
    """Throw error if user is missing required columns"""
    expected_cols = ["County", "State"]
    missing = [col for col in expected_cols if col not in df]
    if missing:
        msg = (
            "The following required columns were not found in the county "
            f"input: {missing}"
        )
        raise COMPASSValueError(msg)


def _filter_not_found_counties(df):
    """Filter out counties with null FIPS codes"""
    _warn_about_missing_counties(df)
    return df[~df.FIPS.isna()].copy()


def _warn_about_missing_counties(df):
    """Throw warning about counties that were not found in the list"""
    not_found_counties = df[df.FIPS.isna()]
    if len(not_found_counties):
        not_found_counties_str = not_found_counties[
            ["County", "State"]
            # cspell: disable-next-line
        ].to_markdown(index=False, tablefmt="psql")
        msg = (
            "The following counties were not found! Please make sure to "
            "use proper spelling and capitalization.\n"
            f"{not_found_counties_str}"
        )
        logger.warning(msg)
        warn(msg)


def _format_county_df_for_output(df):
    """Format county DataFrame for output."""
    out_cols = ["County", "State", "County Type", "FIPS", "Website"]
    df.FIPS = df.FIPS.astype(int)
    return df[out_cols].reset_index(drop=True)


def _convert_to_title(df, column):
    """Convert the values of a DataFrame column to titles."""
    df[column] = df[column].str.strip().str.casefold().str.title()
    return df



"""Ordinance text extraction tooling"""

from .apply import (
    check_for_ordinance_info,
    extract_date,
    extract_ordinance_text_with_llm,
    extract_ordinance_text_with_ngram_validation,
    extract_ordinance_values,
)



"""Wind ordinance extraction utilities"""

from .ordinance import (
    WindHeuristic,
    WindOrdinanceTextCollector,
    WindOrdinanceTextExtractor,
    WindPermittedUseDistrictsTextCollector,
    WindPermittedUseDistrictsTextExtractor,
)
from .parse import (
    StructuredWindOrdinanceParser,
    StructuredWindPermittedUseDistrictsParser,
)


WIND_QUESTION_TEMPLATES = [
    "filetype:pdf {location} wind energy conversion system ordinances",
    "wind energy conversion system ordinances {location}",
    "{location} wind WECS ordinance",
    "Where can I find the legal text for commercial wind energy "
    "conversion system zoning ordinances in {location}?",
    "What is the specific legal information regarding zoning "
    "ordinances for commercial wind energy conversion systems in {location}?",
]



"""Common ordinance extraction components"""

import asyncio
import logging

import networkx as nx
from elm import ApiBase

from compass.extraction.tree import AsyncDecisionTree
from compass.utilities import llm_response_as_json
from compass.utilities.enums import LLMUsageCategory
from compass.utilities.parsing import (
    merge_overlapping_texts,
    clean_backticks_from_llm_response,
)


logger = logging.getLogger(__name__)
_SECTION_PROMPT = (
    'The value of the "section" key should be a string representing the '
    "title of the section (including numerical labels), if it's given, "
    "and `null` otherwise."
)
_SUMMARY_PROMPT = (
    'The value of the "summary" key should be a short summary '
    "of the ordinance, using direct text excerpts as much as possible."
)
_UNITS_IN_SUMMARY_PROMPT = (
    "Include any clarifications about the units in the summary."
)
EXTRACT_ORIGINAL_TEXT_PROMPT = (
    "Extract all portions of the text (with original formatting) "
    "that state how close I can site {tech} to {feature}"
)


def setup_graph_no_nodes(**kwargs):
    """Setup a graph with no nodes

    This function is used to set keywords on the graph that can be used
    in text prompts on the graph nodes.

    Returns
    -------
    nx.DiGraph
        Graph with no nodes but with global keywords set.
    """
    return nx.DiGraph(
        SECTION_PROMPT=_SECTION_PROMPT,
        SUMMARY_PROMPT=_SUMMARY_PROMPT,
        UNITS_IN_SUMMARY_PROMPT=_UNITS_IN_SUMMARY_PROMPT,
        **kwargs,
    )


def llm_response_starts_with_yes(response):
    """Check if LLM response begins with "yes" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "Yes".
    """
    return response.lower().startswith("yes")


def llm_response_starts_with_no(response):
    """Check if LLM response begins with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "No".
    """
    return response.lower().startswith("no")


def llm_response_does_not_start_with_no(response):
    """Check if LLM response does not start with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response does not begin with "No".
    """
    return not llm_response_starts_with_no(response)


def setup_async_decision_tree(graph_setup_func, **kwargs):
    """Setup Async Decision tree for ordinance extraction"""
    G = graph_setup_func(**kwargs)  # noqa: N806
    tree = AsyncDecisionTree(G)
    assert len(tree.chat_llm_caller.messages) == 1
    return tree


def found_ord(messages):
    """Check messages from the base graph to see if ordinance was found

    ..IMPORTANT:: This function may break if the base graph structure
                  changes. Always update the hardcoded values to match
                  the base graph message containing the LLM response
                  about ordinance content.
    """
    num_messages_in_base_tree = 3
    if len(messages) < num_messages_in_base_tree:
        return False
    return llm_response_starts_with_yes(messages[2].get("content", ""))


async def run_async_tree(tree, response_as_json=True):
    """Run Async Decision Tree and return output as dict"""
    try:
        response = await tree.async_run()
    except RuntimeError:
        msg = (
            "    - NOTE: This is not necessarily an error and may just mean "
            "that the text does not have the requested data."
        )
        logger.error(msg)  # noqa: TRY400
        response = None

    if response_as_json:
        return llm_response_as_json(response) if response else {}

    return response


async def run_async_tree_with_bm(tree, base_messages):
    """Run Async Decision Tree from base messages; return dict output"""
    tree.chat_llm_caller.messages = base_messages
    assert len(tree.chat_llm_caller.messages) == len(base_messages)
    return await run_async_tree(tree)


def empty_output(feature):
    """Empty output for a feature (not found in text)"""
    if feature in {"structures", "property line"}:
        return [
            {"feature": f"{feature} (participating)"},
            {"feature": f"{feature} (non-participating)"},
        ]
    return [{"feature": feature}]


def setup_base_setback_graph(**kwargs):
    """Setup graph to get setback ordinance text for a given feature

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Is there text in the following legal document that describes "
            "how far I have to setback {tech} from {feature}? "
            "{feature_clarifications}"
            "Pay extra attention to clarifying text found in parentheses "
            "and footnotes. Begin your response with either 'Yes' or 'No' "
            "and explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge(
        "init", "get_text", condition=llm_response_does_not_start_with_no
    )
    G.add_node("get_text", prompt=EXTRACT_ORIGINAL_TEXT_PROMPT)

    return G


def setup_participating_owner(**kwargs):
    """Setup graph to check for "participating" setbacks for a feature

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to participating owners? Occupying owners "
            "are not participating owners unless explicitly mentioned in the "
            "text. Justify your answer by quoting the raw text directly."
        ),
    )
    G.add_edge("init", "non_part")
    G.add_node(
        "non_part",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to non-participating owners? Non-occupying "
            "owners are not non-participating owners unless explicitly "
            "mentioned in the text. Justify your answer by quoting the raw "
            "text directly."
        ),
    )
    G.add_edge("non_part", "final")
    G.add_node(
        "final",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly two "
            'keys. The keys are "participating" and "non-participating". The '
            'value of the "participating" key should be a string containing '
            "the raw text with original formatting from the ordinance that "
            "applies to participating owners or `null` if there was no such "
            'text. The value of the "non-participating" key should be a '
            "string containing the raw text with original formatting from the "
            "ordinance that applies to non-participating owners or simply the "
            "full ordinance if the text did not make the distinction between "
            "participating and non-participating owners."
        ),
    )
    return G


def setup_graph_extra_restriction(is_numerical=True, **kwargs):
    """Setup Graph to extract non-setback ordinance values from text

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    kwargs.setdefault("unit_clarification", "")
    kwargs.setdefault("feature_clarifications", "")
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the following text explicitly mention {restriction} for "
            "{tech}? {feature_clarifications}\nMake sure your answer adheres "
            "to these guidelines:\n"
            "1) Respond based only on the explicit text provided for "
            "{restriction}. Do not infer values from text based on "
            "related restrictions. If {restriction} is not explicitly "
            "mentioned, then say 'No'.\n"
            "2) Pay close attention to clarifying details in parentheses, "
            "footnotes, or additional explanatory text.\n"
            "3) Begin your response with either 'Yes' or 'No' and explain "
            "your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    if is_numerical:
        G.add_edge("init", "value", condition=llm_response_starts_with_yes)
        G.add_node(
            "value",
            prompt=(
                "What is the **numerical** value given for the {restriction} "
                "for {tech}? Follow these guidelines:\n"
                "1) Extract only the explicit numerical value provided for "
                "the restriction. Do not infer values from related "
                "restrictions.\n"
                "2) If multiple values are given, select the most restrictive "
                "one (i.e., the smallest allowable limit, the lowest maximum, "
                "etc.).\n"
                "3) Pay close attention to clarifying details in parentheses, "
                "footnotes, or additional explanatory text.\n\n"
                "Example Inputs and Outputs:\n"
                'Text: "For all WES there is a limitation of overall height '
                'of 200 feet (including blades)."\n'
                "Output: 200\n"
                'Text: "The noise level of all SES shall be no greater than '
                "thirty-two (32) decibels measured from the nearest property "
                "line. This level may only be exceeded during short-term "
                'events such as utility outages and/or severe wind storms."\n'
                "Output: 32\n"
                'Text: "At no time shall a wind turbine tower, nacelle, or '
                "blade create shadow flicker on any non-participating "
                'landowner property"\n'
                "Output: 0\n"
                "Text: Solar Panels shall not exceed 22'6\" in height. The "
                "height is determined from the ground to the top of the panel "
                "at any angle.\n"
                "Output: 22.5\n"
            ),
        )

        G.add_edge("value", "units")
        G.add_node(
            "units",
            prompt=(
                "What are the units for the {restriction} for {tech}? "
                "Ensure that:\n1) You accurately identify the unit value "
                "associated with the restriction.\n2) The unit is "
                "expressed using standard, conventional unit names (e.g., "
                '"feet", "meters", "acres", "dBA", etc.). {unit_clarification}'
                "\n3) If multiple values are mentioned, return only the units "
                "for the most restrictive value that directly pertains to the "
                "restriction.\n\nExample Inputs and Outputs:\n"
                'Text: "For all WES there is a limitation of overall height '
                'of 200 feet (including blades)."\n'
                'Output: "feet"\n'
                'Text: "The noise level of all SES shall be no greater than '
                "thirty-two (32) decibels measured from the nearest property "
                "line. This level may only be exceeded during short-term "
                'events such as utility outages and/or severe wind storms."\n'
                'Output: "dBA"\n'
                'Text: "At no time shall a wind turbine tower, nacelle, or '
                "blade create shadow flicker on any non-participating "
                'landowner property"\n'
                'Output: "hr/year"\n'
            ),
        )

        G.add_edge("units", "final")
        G.add_node(
            "final",
            prompt=(
                "Please respond based on our entire conversation so far. "
                "Return your answer in "
                "JSON format (not markdown). Your JSON file must include "
                'exactly four keys. The keys are "value", "units", "summary", '
                'and "section". The value of the "value" key '
                "should be a numerical value corresponding to the "
                "{restriction} for {tech}, or `null` if the text does not "
                "mention such a restriction. Use our conversation to fill "
                'out this value. The value of the "units" key should be a '
                "string corresponding to the (standard) units for the "
                "{restriction} allowed for {tech} by the text below, or "
                "`null` if the text does not mention such a restriction. "
                "{SUMMARY_PROMPT} {UNITS_IN_SUMMARY_PROMPT} {SECTION_PROMPT}"
            ),
        )
    else:
        G.add_edge("init", "final", condition=llm_response_starts_with_yes)
        G.add_node(
            "final",
            prompt=(
                "Please respond based on our entire conversation so far. "
                "Return your answer in "
                "JSON format (not markdown). Your JSON file must include "
                'exactly two keys. The keys are "summary" and "section". '
                "{SUMMARY_PROMPT} {SECTION_PROMPT}"
            ),
        )
    return G


def setup_graph_permitted_use_districts(**kwargs):
    """Setup graph to extract permitted use districts for technology

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(**kwargs)  # noqa: N806

    G.add_node(
        "init",
        prompt=(
            "Does the following text explicitly outline districts where "
            "{tech} are permitted as {use_type}? {clarifications}"
            "Pay extra attention to clarifying text found in "
            "parentheses and footnotes. Begin your response with either "
            "'Yes' or 'No' and explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )
    G.add_edge(
        "init", "district_names", condition=llm_response_starts_with_yes
    )

    G.add_node(
        "district_names",
        prompt=(
            "What are all of the district names (and abbreviations if given) "
            "where {tech} are permitted as {use_type}?"
        ),
    )
    G.add_edge("district_names", "final")

    G.add_node(
        "final",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer in "
            "JSON format (not markdown). Your JSON file must include "
            'exactly three keys. The keys are "value", "summary", '
            'and "section". The value of the "value" key '
            "should be a list of all district names (and abbreviations if "
            "given) where {tech} "
            "are permitted as {use_type}, or `null` if the text does not "
            "mention this use type for {tech}. Use our conversation to "
            "fill out this value. {SUMMARY_PROMPT} {SECTION_PROMPT}"
        ),
    )
    return G


class BaseTextExtractor:
    """Base implementation for a text extractor"""

    SYSTEM_MESSAGE = (
        "You extract exact excerpts from a provided legal zoning "
        "regulation text that are directly relevant to the user's request "
        "along with relevant context (such as section headers, for example). "
        "You must preserve all original wording, punctuation, and formatting "
        "without any paraphrasing or modification. If the relevant content "
        "appears within a space-delimited table, return the entire table "
        "exactly as formatted. You must not summarize, rephrase, interpret, "
        "or provide additional commentary. Only return the original text "
        "excerpts as they appear in the source."
    )
    _USAGE_LABEL = LLMUsageCategory.DOCUMENT_ORDINANCE_SUMMARY

    def __init__(self, llm_caller):
        """

        Parameters
        ----------
        llm_caller : compass.llm.LLMCaller
            LLM Caller instance used to extract ordinance info with.
        """
        self.llm_caller = llm_caller

    async def _process(self, text_chunks, instructions, is_valid_chunk):
        """Perform extraction processing"""
        logger.info(
            "Extracting summary text from %d text chunks asynchronously...",
            len(text_chunks),
        )
        logger.debug("Model instructions are:\n%s", instructions)
        outer_task_name = asyncio.current_task().get_name()
        summaries = [
            asyncio.create_task(
                self.llm_caller.call(
                    sys_msg=self.SYSTEM_MESSAGE,
                    content=f"{instructions}\n\n# TEXT #\n\n{chunk}",
                    usage_sub_label=self._USAGE_LABEL,
                ),
                name=outer_task_name,
            )
            for chunk in text_chunks
        ]
        summary_chunks = await asyncio.gather(*summaries)
        summary_chunks = [
            clean_backticks_from_llm_response(chunk)
            for chunk in summary_chunks
            if is_valid_chunk(chunk)
        ]

        text_summary = merge_overlapping_texts(summary_chunks)
        logger.debug(
            "Final summary contains %d tokens",
            ApiBase.count_tokens(
                text_summary,
                model=self.llm_caller.kwargs.get("model", "gpt-4"),
            ),
        )
        return text_summary



"""Ordinance async decision tree"""

import networkx as nx
import logging

from elm.tree import DecisionTree

from compass.utilities.enums import LLMUsageCategory


logger = logging.getLogger(__name__)


class AsyncDecisionTree(DecisionTree):
    """Async class to traverse a directed graph of LLM prompts

    Nodes of this tree are prompts, and edges are transitions between
    prompts based on conditions being met in the LLM response

    Purpose:
        Represent a series of prompts that can be used in sequence to
        extract values of interest from text.
    Responsibilities:
        1. Store all prompts used to extract a particular ordinance
           value from text.
        2. Track relationships between the prompts (i.e. which prompts
           is used first, which prompt is used next depending on the
           output of the previous prompt, etc.) using a directed acyclic
           graph.
    Key Relationships:
        Inherits from :class:`~elm.tree.DecisionTree` to add ``async``
        capabilities. Uses a
        :class:`~compass.llm.calling.ChatLLMCaller` for LLm queries.
    """

    def __init__(self, graph):
        """

        Parameters
        ----------
        graph : nx.DiGraph
            Directed acyclic graph where nodes are LLM prompts and edges
            are logical transitions based on the response. Must have
            high-level graph attribute "chat_llm_caller" which is a
            ChatLLMCaller instance. Nodes should have attribute "prompt"
            which can have {format} named arguments that will be filled
            from the high-level graph attributes. Edges can have
            attribute "condition" that is a callable to be executed on
            the LLM response text. An edge from a node without a
            condition acts as an "else" statement if no other edge
            conditions are satisfied. A single edge from node to node
            does not need a condition.
        """
        self._g = graph
        self._history = []
        assert isinstance(self.graph, nx.DiGraph)
        assert "chat_llm_caller" in self.graph.graph

    @property
    def chat_llm_caller(self):
        """ChatLLMCaller: ChatLLMCaller instance for this tree"""
        return self.graph.graph["chat_llm_caller"]

    @property
    def messages(self):
        """Get a list of the conversation messages with the LLM

        Returns
        -------
        list
        """
        return self.chat_llm_caller.messages

    @property
    def all_messages_txt(self):
        """Get a printout of the full conversation with the LLM

        Returns
        -------
        str
        """
        messages = [
            f"{msg['role'].upper()}: {msg['content']}" for msg in self.messages
        ]
        return "\n\n".join(messages)

    async def async_call_node(self, node0):
        """Call the LLM

        The chat will start with the prompt from the input node and will
        search the successor edges for a valid transition condition.

        Parameters
        ----------
        node0 : str
            Name of node being executed.

        Returns
        -------
        out : str
            Next node or LLM response if at a leaf node.
        """
        prompt = self._prepare_graph_call(node0)
        out = await self.chat_llm_caller.call(
            prompt, usage_sub_label=LLMUsageCategory.DECISION_TREE
        )
        logger.debug(
            "Chat GPT prompt:\n%s\nChat GPT response:\n%s", prompt, out
        )
        return self._parse_graph_output(node0, out or "")

    async def async_run(self, node0="init"):
        """Traverse the decision tree starting at the input node

        Parameters
        ----------
        node0 : str
            Name of starting node in the graph. This is typically called
            "init".

        Returns
        -------
        out : str
            Final response from LLM at the leaf node.
        """

        self._history = []

        while True:
            try:
                out = await self.async_call_node(node0)
            except Exception as e:
                logger.debug(
                    "Error traversing trees, here's the full "
                    "conversation printout:\n%s",
                    self.all_messages_txt,
                )
                last_message = self.messages[-1]["content"]
                msg = (
                    "Ran into an exception when traversing tree. "
                    "Last message from LLM is printed below. "
                    "See debug logs for more detail. "
                    "\nLast message: \n"
                    '"""\n%s\n"""'
                )
                logger.exception(msg, last_message)
                raise RuntimeError(msg % last_message) from e
            if out in self.graph:
                node0 = out
            else:
                break

        logger.info("Output: %s", out)

        return out



"""Solar ordinance extraction utilities"""

from .ordinance import (
    SolarHeuristic,
    SolarOrdinanceTextCollector,
    SolarOrdinanceTextExtractor,
    SolarPermittedUseDistrictsTextCollector,
    SolarPermittedUseDistrictsTextExtractor,
)
from .parse import (
    StructuredSolarOrdinanceParser,
    StructuredSolarPermittedUseDistrictsParser,
)


SOLAR_QUESTION_TEMPLATES = [
    "filetype:pdf {location} solar energy conversion system ordinances",
    "solar energy conversion system ordinances {location}",
    "{location} solar energy farm ordinance",
    "Where can I find the legal text for commercial solar energy "
    "conversion system zoning ordinances in {location}?",
    "What is the specific legal information regarding zoning "
    "ordinances for commercial solar energy conversion systems in {location}?",
]



"""COMPASS enum definitions"""

from enum import StrEnum, auto


class LLMUsageCategory(StrEnum):
    """COMPASS LLM usage categories"""

    CHAT = auto()
    DATE_EXTRACTION = auto()
    DECISION_TREE = auto()
    DEFAULT = auto()
    DOCUMENT_CONTENT_VALIDATION = auto()
    DOCUMENT_ORDINANCE_SUMMARY = auto()
    DOCUMENT_PERMITTED_USE_CONTENT_VALIDATION = auto()
    DOCUMENT_PERMITTED_USE_DISTRICTS_SUMMARY = auto()
    DOCUMENT_LOCATION_VALIDATION = auto()


class LLMTasks(StrEnum):
    """LLM-based COMPASS tasks"""

    DATE_EXTRACTION = LLMUsageCategory.DATE_EXTRACTION
    """Date extraction task"""

    DEFAULT = LLMUsageCategory.DEFAULT
    """Default fallback option for all tasks"""

    DOCUMENT_CONTENT_VALIDATION = LLMUsageCategory.DOCUMENT_CONTENT_VALIDATION
    """Document content validation task

    This represents a task like "does the document contain ordinance
    values" or "does the document contain permitted use specifications".
    """

    DOCUMENT_LOCATION_VALIDATION = (
        LLMUsageCategory.DOCUMENT_LOCATION_VALIDATION
    )
    """Document location validation task

    This represents all the tasks associated with validation that the
    document pertains to a particular jurisdiction.
    """

    ORDINANCE_TEXT_EXTRACTION = auto()
    """Ordinance text extraction task

    This task represents the extraction/summarization of text containing
    ordinance values.
    """

    PERMITTED_USE_TEXT_EXTRACTION = auto()
    """Permitted use text extraction task

    This task represents the extraction/summarization of text containing
    permitted use descriptions and allowances.
    """

    ORDINANCE_VALUE_EXTRACTION = auto()
    """Ordinance structured value extraction task

    This task represents the extraction of structured ordinance values.
    """

    PERMITTED_USE_VALUE_EXTRACTION = auto()
    """Permitted use structured value extraction task

    This task represents the extraction of structured permitted use
    values.
    """



"""Ordinance county file downloading logic"""

import logging

from elm.web.document import PDFDocument
from elm.web.search.run import web_search_links_as_docs
from elm.web.utilities import filter_documents

from compass.llm import StructuredLLMCaller
from compass.extraction import check_for_ordinance_info, extract_date
from compass.services.threaded import TempFileCache
from compass.validation.location import (
    CountyJurisdictionValidator,
    CountyNameValidator,
    CountyValidator,
)
from compass.utilities.enums import LLMTasks
from compass.pb import COMPASS_PB


logger = logging.getLogger(__name__)


async def download_county_ordinance(
    question_templates,
    location,
    model_configs,
    heuristic,
    ordinance_text_collector_class,
    permitted_use_text_collector_class,
    num_urls=5,
    file_loader_kwargs=None,
    browser_semaphore=None,
    usage_tracker=None,
):
    """Download the ordinance document(s) for a single county

    Parameters
    ----------
    location : :class:`compass.utilities.location.Location`
        Location objects representing the county.
    model_configs : dict
        Dictionary of :class:`~compass.llm.config.LLMConfig` instances.
        Should have at minium a "default" key that is used as a fallback
        for all tasks.
    num_urls : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader` with. If found, the
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.search.google.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    browser_semaphore : :class:`asyncio.Semaphore`, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    usage_tracker : compass.services.usage.UsageTracker, optional
        Optional tracker instance to monitor token usage during
        LLM calls. By default, ``None``.

    Returns
    -------
    list or None
        List of :obj:`~elm.web.document.BaseDocument` instances possibly
        containing ordinance information, or ``None`` if no ordinance
        document was found.
    """
    COMPASS_PB.update_jurisdiction_task(
        location.full_name, description="Downloading files..."
    )
    docs = await _docs_from_web_search(
        question_templates,
        location,
        num_urls,
        browser_semaphore,
        **(file_loader_kwargs or {}),
    )
    COMPASS_PB.update_jurisdiction_task(
        location.full_name,
        description="Checking files for correct jurisdiction...",
    )
    docs = await _down_select_docs_correct_location(
        docs,
        location=location,
        usage_tracker=usage_tracker,
        model_config=model_configs.get(
            LLMTasks.DOCUMENT_LOCATION_VALIDATION,
            model_configs[LLMTasks.DEFAULT],
        ),
    )
    logger.info(
        "%d document(s) remaining after location filter for %s\n\t- %s",
        len(docs),
        location.full_name,
        "\n\t- ".join(
            [doc.attrs.get("source", "Unknown source") for doc in docs]
        ),
    )
    COMPASS_PB.update_jurisdiction_task(
        location.full_name, description="Checking files for legal text..."
    )
    docs = await _down_select_docs_correct_content(
        docs,
        location=location,
        model_configs=model_configs,
        heuristic=heuristic,
        ordinance_text_collector_class=ordinance_text_collector_class,
        permitted_use_text_collector_class=permitted_use_text_collector_class,
        usage_tracker=usage_tracker,
    )
    logger.info(
        "Found %d potential ordinance documents for %s\n\t- %s",
        len(docs),
        location.full_name,
        "\n\t- ".join(
            [doc.attrs.get("source", "Unknown source") for doc in docs]
        ),
    )
    return _sort_final_ord_docs(docs)


async def _docs_from_web_search(
    question_templates,
    location,
    num_urls,
    browser_semaphore,
    **file_loader_kwargs,
):
    """Download docs from web using location queries"""
    queries = [
        question.format(location=location.full_name)
        for question in question_templates
    ]
    file_loader_kwargs.update({"file_cache_coroutine": TempFileCache.call})
    return await web_search_links_as_docs(
        queries,
        num_urls=num_urls,
        browser_semaphore=browser_semaphore,
        task_name=location.full_name,
        **file_loader_kwargs,
    )


async def _down_select_docs_correct_location(
    docs, location, usage_tracker, model_config
):
    """Remove all documents not pertaining to the location"""
    llm_caller = StructuredLLMCaller(
        llm_service=model_config.llm_service,
        usage_tracker=usage_tracker,
        **model_config.llm_call_kwargs,
    )
    county_validator = CountyValidator(
        llm_caller, text_splitter=model_config.text_splitter
    )
    return await filter_documents(
        docs,
        validation_coroutine=county_validator.check,
        task_name=location.full_name,
        county=location.name,
        state=location.state,
    )


async def _down_select_docs_correct_content(
    docs,
    location,
    model_configs,
    heuristic,
    ordinance_text_collector_class,
    permitted_use_text_collector_class,
    usage_tracker,
):
    """Remove all documents that don't contain ordinance info"""
    return await filter_documents(
        docs,
        validation_coroutine=_contains_ordinances,
        task_name=location.full_name,
        model_configs=model_configs,
        heuristic=heuristic,
        ordinance_text_collector_class=ordinance_text_collector_class,
        permitted_use_text_collector_class=permitted_use_text_collector_class,
        usage_tracker=usage_tracker,
    )


async def _contains_ordinances(
    doc, model_configs, usage_tracker=None, **kwargs
):
    """Helper coroutine that checks for ordinance and date info"""
    model_config = model_configs.get(
        LLMTasks.DOCUMENT_CONTENT_VALIDATION,
        model_configs[LLMTasks.DEFAULT],
    )
    doc = await check_for_ordinance_info(
        doc,
        model_config=model_config,
        usage_tracker=usage_tracker,
        **kwargs,
    )
    contains_ordinances = doc.attrs.get("contains_ord_info", False)
    if contains_ordinances:
        logger.debug("Detected ordinance info; parsing date...")
        date_model_config = model_configs.get(
            LLMTasks.DATE_EXTRACTION, model_configs[LLMTasks.DEFAULT]
        )
        doc = await extract_date(
            doc, date_model_config, usage_tracker=usage_tracker
        )
    return contains_ordinances


def _sort_final_ord_docs(all_ord_docs):
    """Sort the list of documents by year, type, and text length"""
    if not all_ord_docs:
        return None

    return sorted(all_ord_docs, key=_ord_doc_sorting_key, reverse=True)


def _ord_doc_sorting_key(doc):
    """Sorting key for documents. The higher this value, the better"""
    latest_year, latest_month, latest_day = doc.attrs.get("date", (-1, -1, -1))
    prefer_pdf_files = isinstance(doc, PDFDocument)
    highest_name_score = doc.attrs.get(
        # missing key means we were so confident that check wasn't
        # even applied, so we default to 1 here
        CountyNameValidator.META_SCORE_KEY,
        1,
    )
    highest_jurisdiction_score = doc.attrs.get(
        CountyJurisdictionValidator.META_SCORE_KEY, 0
    )
    shortest_text_length = -1 * len(doc.text)
    return (
        latest_year,
        prefer_pdf_files,
        highest_name_score,
        highest_jurisdiction_score,
        shortest_text_length,
        latest_month,
        latest_day,
    )



"""Ordinance function to apply ordinance extraction on a document"""

import logging
from warnings import warn

from compass.llm import StructuredLLMCaller
from compass.extraction.date import DateExtractor
from compass.validation import (
    ParseChunksWithMemory,
    LegalTextValidator,
    parse_by_chunks,
)
from compass.warn import COMPASSWarning


logger = logging.getLogger(__name__)


async def check_for_ordinance_info(
    doc,
    model_config,
    heuristic,
    ordinance_text_collector_class,
    permitted_use_text_collector_class=None,
    usage_tracker=None,
):
    """Parse a single document for ordinance information

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document potentially containing ordinance information. Note
        that if the document's attrs contains the
        ``"contains_ord_info"`` key, it will not be processed. To force
        a document to be processed by this function, remove that key
        from the documents attrs.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    usage_tracker : compass.services.usage.UsageTracker, optional
        Optional tracker instance to monitor token usage during
        LLM calls. By default, ``None``.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the parsing are stored in the documents attrs. In particular,
        the attrs will contain a ``"contains_ord_info"`` key that
        will be set to ``True`` if ordinance info was found in the text,
        and ``False`` otherwise. If ``True``, the attrs will also
        contain a ``"date"`` key containing the most recent date that
        the ordinance was enacted (or a tuple of `None` if not found),
        and an ``"ordinance_text"`` key containing the ordinance text
        snippet. Note that the snippet may contain other info as well,
        but should encapsulate all of the ordinance text.
    """
    if "contains_ord_info" in doc.attrs:
        return doc

    llm_caller = StructuredLLMCaller(
        llm_service=model_config.llm_service,
        usage_tracker=usage_tracker,
        **model_config.llm_call_kwargs,
    )
    chunks = model_config.text_splitter.split_text(doc.text)
    chunk_parser = ParseChunksWithMemory(llm_caller, chunks, num_to_recall=2)
    legal_text_validator = LegalTextValidator()

    ordinance_text_collector = ordinance_text_collector_class()
    callbacks = [ordinance_text_collector.check_chunk]
    if permitted_use_text_collector_class is not None:
        permitted_use_text_collector = permitted_use_text_collector_class()
        callbacks.append(permitted_use_text_collector.check_chunk)

    doc.attrs["is_legal_text"] = await parse_by_chunks(
        chunk_parser,
        heuristic,
        legal_text_validator,
        callbacks=callbacks,
        min_chunks_to_process=3,
    )

    doc.attrs["contains_ord_info"] = ordinance_text_collector.contains_ord_info
    if doc.attrs["contains_ord_info"]:
        doc.attrs["ordinance_text"] = ordinance_text_collector.ordinance_text
        logger.debug(
            "Ordinance text for %s is:\n%s",
            doc.attrs.get("source", "unknown source"),
            doc.attrs["ordinance_text"],
        )

    if permitted_use_text_collector_class is not None:
        doc.attrs["contains_district_info"] = (
            permitted_use_text_collector.contains_district_info
        )
        if doc.attrs["contains_district_info"]:
            doc.attrs["permitted_use_text"] = (
                permitted_use_text_collector.permitted_use_district_text
            )
            logger.debug(
                "Permitted use text for %s is:\n%s",
                doc.attrs.get("source", "unknown source"),
                doc.attrs["permitted_use_text"],
            )

    return doc


async def extract_date(doc, model_config, usage_tracker=None):
    """Parse a single document for date information

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document potentially containing date information.
    usage_tracker : compass.services.usage.UsageTracker, optional
        Optional tracker instance to monitor token usage during
        LLM calls. By default, ``None``.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for dates. The results of
        the parsing are stored in the documents attrs. In particular,
        the attrs will contain a ``"date"`` key that will contain the
        parsed date information.
    """
    date_llm_caller = StructuredLLMCaller(
        llm_service=model_config.llm_service,
        usage_tracker=usage_tracker,
        **model_config.llm_call_kwargs,
    )
    doc.attrs["date"] = await DateExtractor(
        date_llm_caller, model_config.text_splitter
    ).parse(doc)

    return doc


async def extract_ordinance_text_with_llm(
    doc, text_splitter, extractor, original_text_key
):
    """Extract ordinance text from document using LLM

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance information. This means it
        must contain an ``"ordinance_text"`` key in the attrs. You can
        run :func:`~compass.extraction.apply.check_for_ordinance_info`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's attrs does not contain the ``"ordinance_text"``
        key, you will get an error.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    extractor : compass.extraction.ordinance.WindOrdinanceTextExtractor
        Instance of
        :class:`~compass.extraction.ordinance.WindOrdinanceTextExtractor`
        used for ordinance text extraction.
    original_text_key : str
        String corresponding to the `doc.attrs` key containing the
        original text (before extraction).

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the extraction are stored in the document's attrs.
    str
        Key corresponding to the cleaned ordinance text stored in the
        `doc.attrs` dictionary.

    """
    prev_meta_name = original_text_key  # "ordinance_text"
    for meta_name, parser in extractor.parsers:
        doc.attrs[meta_name] = await _parse_if_input_text_not_empty(
            doc.attrs[prev_meta_name],
            text_splitter,
            parser,
            prev_meta_name,
            meta_name,
        )
        prev_meta_name = meta_name

    return doc, prev_meta_name


async def extract_ordinance_text_with_ngram_validation(
    doc,
    text_splitter,
    extractor,
    original_text_key,
    n=4,
    num_extraction_attempts=3,
    ngram_fraction_threshold=0.75,
):
    """Extract ordinance text for a single document with known ord info

    This extraction includes an "ngram" check, which attempts to detect
    whether or not the cleaned text was extracted from the original
    ordinance text. The processing will attempt to re-extract the text
    if the validation does not pass a certain threshold until the
    maximum number of attempts is reached. If the text still does not
    pass validation at this point, there is a good chance that the LLM
    hallucinated parts of the output text, so caution should be taken.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance information. This means it
        must contain an ``"ordinance_text"`` key in the attrs. You can
        run :func:`~compass.extraction.apply.check_for_ordinance_info`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's attrs does not contain the ``"ordinance_text"``
        key, it will not be processed.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    original_text_key : str
        String corresponding to the `doc.attrs` key containing the
        original text (before extraction).
    n : int, optional
        Number of words to include per ngram for the ngram validation,
        which helps ensure that the LLM did not hallucinate.
        By default, ``4``.
    num_extraction_attempts : int, optional
        Number of extraction attempts before returning text that did not
        pass the ngram check. If the processing exceeds this value,
        there is a good chance that the LLM hallucinated parts of the
        output text. Cannot be negative or 0. By default, ``3``.
    ngram_fraction_threshold : float, optional
        Fraction of ngrams in the cleaned text that are also found in
        the original ordinance text for the extraction to be considered
        successful. Should be a value between 0 and 1 (inclusive).
        By default, ``0.95``.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the extraction are stored in the document's attrs.
    """
    if not doc.attrs.get(original_text_key):
        msg = (
            f"Input document has no {original_text_key!r} key or string "
            "does not contain information. Please run "
            "`check_for_ordinance_info` prior to calling this method."
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    return await _extract_with_ngram_check(
        doc,
        text_splitter,
        extractor,
        original_text_key,
        n=max(1, n),
        num_tries=max(1, num_extraction_attempts),
        ngram_fraction_threshold=max(0, min(1, ngram_fraction_threshold)),
    )


async def _extract_with_ngram_check(
    doc,
    text_splitter,
    extractor,
    original_text_key,
    n=4,
    num_tries=3,
    ngram_fraction_threshold=0.75,
):
    """Extract ordinance info from doc and validate using ngrams."""
    from compass.extraction.ngrams import sentence_ngram_containment  # noqa

    source = doc.attrs.get("source", "Unknown")
    original_text = doc.attrs[original_text_key]
    if not original_text:
        msg = (
            "Document missing original ordinance text! No extraction "
            "performed (Document source: %s)",
            source,
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    best_score = 0
    best_summary = ""
    out_text_key = "extracted_text"
    for attempt in range(1, num_tries + 1):
        doc, out_text_key = await extract_ordinance_text_with_llm(
            doc, text_splitter, extractor, original_text_key
        )
        cleaned_text = doc.attrs[out_text_key]
        if not cleaned_text:
            logger.debug(
                "No cleaned text found after extraction on attempt %d "
                "for document with source %s. Retrying...",
                attempt,
                source,
            )
            continue

        ngram_frac = sentence_ngram_containment(
            original=original_text, test=cleaned_text, n=n
        )
        if ngram_frac >= ngram_fraction_threshold:
            logger.debug(
                "Document extraction for %r passed ngram check on attempt %d "
                "with score %.2f (Document source: %s)",
                out_text_key,
                attempt + 1,
                ngram_frac,
                source,
            )
            best_score = ngram_frac
            break

        if ngram_frac > best_score:
            best_score = ngram_frac
            best_summary = cleaned_text

        logger.debug(
            "Document extraction for %r failed ngram check on attempt %d "
            "with score %.2f (Document source: %s). Retrying...",
            out_text_key,
            attempt + 1,
            ngram_frac,
            source,
        )
    else:
        doc.attrs[out_text_key] = best_summary
        msg = (
            f"Ngram check failed after {num_tries} tries. LLM hallucination "
            "in cleaned ordinance text is extremely likely! Proceed with "
            f"caution!! (Score: {best_score:.2f}; Document source: {source})"
        )
        logger.warning(msg)
        warn(msg, UserWarning)

    doc.attrs["ngram_score"] = best_score
    return doc


async def extract_ordinance_values(doc, parser, text_key, out_key):
    """Extract ordinance values for a single document

    Document must be known to contain ordinance text.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance text. This means it must
        contain an `text_key` key in the attrs. You can run
        :func:`~compass.extraction.apply.extract_ordinance_text_with_llm`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's attrs does not contain the `text_key` key, it will
        not be processed.
    text_key : str
        Name of the key under which cleaned text is stored in
        `doc.attrs`. This text should be ready for extraction.
    out_key : str
        Name of the key under which extracted ordinances should be
        stored.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance values. The results
        of the extraction are stored in the document's attrs.
    """
    if not doc.attrs.get(text_key):
        msg = (
            f"Input document has no {text_key!r} key or string "
            "does not contain info. Please run "
            "`extract_ordinance_text_with_llm` prior to calling this method."
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    doc.attrs[out_key] = await parser.parse(doc.attrs[text_key])
    return doc


async def _parse_if_input_text_not_empty(
    text, text_splitter, parser, curr_text_name, next_text_name
):
    """Extract text using parser, or return empty if input empty"""
    if not text:
        msg = (
            f"{curr_text_name!r} does not contain any text. Skipping "
            f"extraction for {next_text_name!r}"
        )
        warn(msg, COMPASSWarning)
        return text

    text_chunks = text_splitter.split_text(text)
    extracted_text = await parser(text_chunks)

    logger.debug(
        "Extracted text for %r is:\n%s", next_text_name, extracted_text
    )
    return extracted_text



"""COMPASS abstract Service class"""

import asyncio
import logging
from abc import ABC, abstractmethod

from compass.services.queues import get_service_queue
from compass.exceptions import COMPASSNotInitializedError


logger = logging.getLogger(__name__)
MISSING_SERVICE_MESSAGE = """Must initialize the queue for {service_name!r}.
You can likely use the following code structure to fix this:

    from compass.services.provider import RunningAsyncServices

    services = [
        ...
        {service_name}(...),
        ...
    ]
    async with RunningAsyncServices(services):
        # function call here

"""


class Service(ABC):
    """Abstract base class for a Service that can be queued to run"""

    MAX_CONCURRENT_JOBS = 10_000
    """Max number of concurrent job submissions."""

    @classmethod
    def _queue(cls):
        """Get queue for class."""
        service_name = cls.__name__
        queue = get_service_queue(service_name)
        if queue is None:
            msg = MISSING_SERVICE_MESSAGE.format(service_name=service_name)
            raise COMPASSNotInitializedError(msg)
        return queue

    @classmethod
    async def call(cls, *args, **kwargs):
        """Call the service

        Parameters
        ----------
        *args, **kwargs
            Positional and keyword arguments to be passed to the
            underlying service processing function.

        Returns
        -------
        obj
            A response object from the underlying service.
        """
        fut = asyncio.Future()
        outer_task_name = asyncio.current_task().get_name()
        await cls._queue().put((fut, outer_task_name, args, kwargs))
        return await fut

    @property
    def name(self):
        """str: Service name used to pull the correct queue object"""
        return self.__class__.__name__

    async def process_using_futures(self, fut, *args, **kwargs):
        """Process a call to the service

        Parameters
        ----------
        fut : asyncio.Future
            A future object that should get the result of the processing
            operation. If the processing function returns ``answer``,
            this method should call ``fut.set_result(answer)``.
        **kwargs
            Keyword arguments to be passed to the
            underlying processing function.
        """

        try:
            response = await self.process(*args, **kwargs)
        except Exception as e:  # noqa: BLE001
            fut.set_exception(e)
            return

        fut.set_result(response)

    def acquire_resources(self):  # noqa: B027
        """Use this method to allocate resources, if needed"""

    def release_resources(self):  # noqa: B027
        """Use this method to clean up resources, if needed"""

    @property
    @abstractmethod
    def can_process(self):
        """Check if process function can be called.

        This should be a fast-running method that returns a boolean
        indicating whether or not the service can accept more
        processing calls.
        """

    @abstractmethod
    async def process(self, *args, **kwargs):
        """Process a call to the service.

        Parameters
        ----------
        *args, **kwargs
            Positional and keyword arguments to be passed to the
            underlying processing function.
        """


class LLMService(Service):
    """Base class for LLm service

    This service differs from other services in that it must be used
    as an object, not as a class. that is, users must initialize it and
    pass it around in functions in order to use it.
    """

    def __init__(self, model_name, rate_limit, rate_tracker, service_tag=None):
        """

        Parameters
        ----------
        model_name : str
            Name of model being used.
        rate_limit : int or float
            Max usage per duration of the rate tracker. For example,
            if the rate tracker is set to compute the total over
            minute-long intervals, this value should be the max usage
            per minute.
        rate_tracker : `TimeBoundedUsageTracker`
            A TimeBoundedUsageTracker instance. This will be used to
            track usage per time interval and compare to `rate_limit`.
        service_tag : str, optional
            optional tag to use to distinguish service (i.e. make unique
            from other services). Must set this if multiple models with
            the same name are run concurrently. By default, ``None``.
        """
        self.model_name = model_name
        self.rate_limit = rate_limit
        self.rate_tracker = rate_tracker
        self.service_tag = service_tag or ""

    @property
    def can_process(self):
        """bool: Check if usage is under the rate limit"""
        return self.rate_tracker.total < self.rate_limit

    @property
    def name(self):
        """str: Unique service name used to pull the correct queue"""
        return f"{self.__class__.__name__}-{self.model_name}{self.service_tag}"

    def _queue(self):
        """Get queue for class"""
        queue = get_service_queue(self.name)
        if queue is None:
            msg = MISSING_SERVICE_MESSAGE.format(service_name=self.name)
            raise COMPASSNotInitializedError(msg)
        return queue

    async def call(self, *args, **kwargs):
        """Call the service

        Parameters
        ----------
        *args, **kwargs
            Positional and keyword arguments to be passed to the
            underlying service processing function.

        Returns
        -------
        obj
            A response object from the underlying service.
        """
        fut = asyncio.Future()
        outer_task_name = asyncio.current_task().get_name()
        await self._queue().put((fut, outer_task_name, args, kwargs))
        return await fut



"""Ordinance document content Validation logic

These are primarily used to validate that a legal document applies to a
particular technology (e.g. Large Wind Energy Conversion Systems).
"""

import asyncio
import logging
from abc import ABC, abstractmethod

from compass.utilities.enums import LLMUsageCategory


logger = logging.getLogger(__name__)


class ParseChunksWithMemory:
    """Check text chunks by sometimes looking at previous chunks

    The idea behind this approach is that sometimes the context for a
    setback or other ordinances is found in a previous chunk, so it may
    be worthwhile (especially for validation purposes) to check a few
    text chunks back for some validation pieces. In order to do this
    semi-efficiently, we make use of a cache that's labeled "memory".
    """

    def __init__(self, structured_llm_caller, text_chunks, num_to_recall=2):
        """

        Parameters
        ----------
        structured_llm_caller : compass.llm.StructuredLLMCaller
            StructuredLLMCaller instance. Used for structured validation
            queries.
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks. This validator may refer to previous text chunks to
            answer validation questions.
        num_to_recall : int, optional
            Number of chunks to check for each validation call. This
            includes the original chunk! For example, if
            `num_to_recall=2`, the validator will first check the chunk
            at the requested index, and then the previous chunk as well.
            By default, ``2``.
        """
        self.slc = structured_llm_caller
        self.text_chunks = text_chunks
        self.num_to_recall = num_to_recall
        self.memory = [{} for _ in text_chunks]

    # fmt: off
    def _inverted_mem(self, starting_ind):
        """Inverted memory"""
        inverted_mem = self.memory[:starting_ind + 1:][::-1]
        yield from inverted_mem[:self.num_to_recall]

    # fmt: off
    def _inverted_text(self, starting_ind):
        """Inverted text chunks"""
        inverted_text = self.text_chunks[:starting_ind + 1:][::-1]
        yield from inverted_text[:self.num_to_recall]

    async def parse_from_ind(self, ind, prompt, key):
        """Validate a chunk of text

        Validation occurs by querying the LLM using the input prompt and
        parsing the `key` from the response JSON. The prompt should
        request that the key be a boolean output. If the key retrieved
        from the LLM response is False, a number of previous text chunks
        are checked as well, using the same prompt. This can be helpful
        in cases where the answer to the validation prompt (e.g. does
        this text pertain to a large WECS?) is only found in a previous
        text chunk.

        Parameters
        ----------
        ind : int
            Positive integer corresponding to the chunk index.
            Must be less than `len(text_chunks)`.
        prompt : str
            Input LLM system prompt that describes the validation
            question. This should request a JSON output from the LLM.
            It should also take `key` as a formatting input.
        key : str
            A key expected in the JSON output of the LLM containing the
            response for the validation question. This string will also
            be used to format the system prompt before it is passed to
            the LLM.

        Returns
        -------
        bool
            ``True`` if the LLM returned ``True`` for this text chunk or
            `num_to_recall-1` text chunks before it.
            ``False`` otherwise.
        """
        logger.debug("Checking %r for ind %d", key, ind)
        mem_text = zip(
            self._inverted_mem(ind), self._inverted_text(ind), strict=False
        )
        for step, (mem, text) in enumerate(mem_text):
            logger.debug("Mem at ind %d is %s", step, mem)
            check = mem.get(key)
            if check is None:
                content = await self.slc.call(
                    sys_msg=prompt.format(key=key),
                    content=text,
                    usage_sub_label=(
                        LLMUsageCategory.DOCUMENT_CONTENT_VALIDATION
                    ),
                )
                logger.debug("LLM response: %s", str(content))  # TODO: trace
                check = mem[key] = content.get(key, False)
            if check:
                return check
        return False


class Heuristic(ABC):
    """Perform a heuristic check for mention of a technology in text"""

    _GOOD_ACRONYM_CONTEXTS = [
        " {acronym} ",
        " {acronym}\n",
        " {acronym}.",
        "\n{acronym} ",
        "\n{acronym}.",
        "\n{acronym}\n",
        "({acronym} ",
        " {acronym})",
    ]

    def check(self, text, match_count_threshold=1):
        """Check for mention of a tech in text

        This check first strips the text of any tech "look-alike" words
        (e.g. "window", "windshield", etc for "wind" technology). Then,
        it checks for particular keywords, acronyms, and phrases that
        pertain to the tech in the text. If enough keywords are mentions
        (as dictated by `match_count_threshold`), this check returns
        ``True``.

        Parameters
        ----------
        text : str
            Input text that may or may not mention the technology of
            interest.
        match_count_threshold : int, optional
            Number of keywords that must match for the text to pass this
            heuristic check. Count must be strictly greater than this
            value. By default, ``1``.

        Returns
        -------
        bool
            ``True`` if the number of keywords/acronyms/phrases detected
            exceeds the `match_count_threshold`.
        """
        heuristics_text = self._convert_to_heuristics_text(text)
        total_keyword_matches = self._count_single_keyword_matches(
            heuristics_text
        )
        total_keyword_matches += self._count_acronym_matches(heuristics_text)
        total_keyword_matches += self._count_phrase_matches(heuristics_text)
        return total_keyword_matches > match_count_threshold

    def _convert_to_heuristics_text(self, text):
        """Convert text for heuristic content parsing"""
        heuristics_text = text.casefold()
        for word in self.NOT_TECH_WORDS:
            heuristics_text = heuristics_text.replace(word, "")
        return heuristics_text

    def _count_single_keyword_matches(self, heuristics_text):
        """Count number of good tech keywords that appear in text"""
        return sum(
            keyword in heuristics_text for keyword in self.GOOD_TECH_KEYWORDS
        )

    def _count_acronym_matches(self, heuristics_text):
        """Count number of good tech acronyms that appear in text"""
        acronym_matches = 0
        for context in self._GOOD_ACRONYM_CONTEXTS:
            acronym_keywords = {
                context.format(acronym=acronym)
                for acronym in self.GOOD_TECH_ACRONYMS
            }
            acronym_matches = sum(
                keyword in heuristics_text for keyword in acronym_keywords
            )
            if acronym_matches > 0:
                break
        return acronym_matches

    def _count_phrase_matches(self, heuristics_text):
        """Count number of good tech phrases that appear in text"""
        return sum(
            all(keyword in heuristics_text for keyword in phrase.split(" "))
            for phrase in self.GOOD_TECH_PHRASES
        )

    @property
    @abstractmethod
    def NOT_TECH_WORDS(self):  # noqa: N802
        """iter: Iterable of words that don't pertain to the tech"""
        raise NotImplementedError

    @property
    @abstractmethod
    def GOOD_TECH_KEYWORDS(self):  # noqa: N802
        """iter: Iterable of keywords that pertain to the tech"""
        raise NotImplementedError

    @property
    @abstractmethod
    def GOOD_TECH_ACRONYMS(self):  # noqa: N802
        """iter: Iterable of acronyms that pertain to the tech"""
        raise NotImplementedError

    @property
    @abstractmethod
    def GOOD_TECH_PHRASES(self):  # noqa: N802
        """iter: Iterable of phrases that pertain to the tech"""
        raise NotImplementedError


class LegalTextValidator:
    """Parse chunks to determine if they contain legal text"""

    IS_LEGAL_TEXT_PROMPT = (
        "# CONTEXT #\n"
        "You are an AI designed to classify text excerpts based on their "
        "source type. The goal is to identify **legally binding regulations "
        "(such as zoning ordinances or enforceable bans)** and filter out "
        "text that does not belong to an in-effect legal statute.\n"
        "Legally binding regulations are formal laws enacted by a governing "
        "authority, containing enforceable rules or restrictions. However, "
        "many documents discuss, summarize, or propose zoning rules without "
        "having legal force. Examples of non-binding documents include, but "
        "are not limited to, **news articles, reports, draft ordinances, "
        "model ordinances, public notices, surveys, development plans, "
        "project-specific plans, conservation plans, summaries, etc.**\n"
        "\n# OBJECTIVE #\n"
        "Your task is to analyze a given text excerpt and determine whether "
        "it belongs to a **legally binding regulation.** If the text is from "
        "any other type of document, it should be classified accordingly and "
        "excluded from legally binding regulations.\n"
        "\n# RESPONSE #\n"
        "Return the classification in **JSON format** with exactly three "
        "keys:\n\n"
        '1. **"summary"** (string) - A concise summary of the text.\n'
        '2. **"type"** (string) - The best-fitting category of the text.\n'
        '3. **"{key}"** (boolean) -\n'
        "\t- `true` if the text is a **legally binding regulation**.\n"
        "\t- `false` if the text belongs to any other type of document.\n\n"
        "Ensure precise classification by prioritizing **legal "
        "enforceability** over content similarity."
    )

    def __init__(self):
        self._legal_text_mem = []

    @property
    def is_legal_text(self):
        """bool: ``True`` if text was found to be from a legal source"""
        if not self._legal_text_mem:
            return False
        return sum(self._legal_text_mem) >= 0.5 * len(self._legal_text_mem)

    async def check_chunk(self, chunk_parser, ind):
        """Check a chunk at a given ind to see if it contains legal text

        Parameters
        ----------
        chunk_parser : ParseChunksWithMemory
            Instance of `ParseChunksWithMemory` that contains a
            `parse_from_ind` method.
        ind : int
            Index of the chunk to check.

        Returns
        -------
        bool
            Boolean flag indicating whether or not the text in the chunk
            resembles legal text.
        """
        is_legal_text = await chunk_parser.parse_from_ind(
            ind, self.IS_LEGAL_TEXT_PROMPT, key="legal_text"
        )
        self._legal_text_mem.append(is_legal_text)
        if is_legal_text:
            logger.debug("Text at ind %d is legal text", ind)
        else:
            logger.debug("Text at ind %d is not legal text", ind)
        return is_legal_text


async def parse_by_chunks(
    chunk_parser,
    heuristic,
    legal_text_validator,
    callbacks=None,
    min_chunks_to_process=3,
):
    """Parse text by chunks, passing to callbacks if it's legal text

    This method goes through the chunks one by one, and passes them to
    the callback parsers if the `legal_text_validator` check passes. If
    `min_chunks_to_process` number of chunks fail the legal text check,
    parsing is aborted.

    Parameters
    ----------
    chunk_parser : ParseChunksWithMemory
        Instance of `ParseChunksWithMemory` that contains the attributes
        `text_chunks` and `num_to_recall`. The chunks in the
        `text_chunks` attribute will be iterated over.
    heuristic : Heuristic
        Instance of `Heuristic` with a `check` method. This should be a
        fast check meant to quickly dispose of chunks of text. Any chunk
        that fails this check will NOT be passed to the callback
        parsers.
    legal_text_validator : LegalTextValidator
        Instance of `LegalTextValidator` that can be used to validate
        each chunk for legal text.
    callbacks : list, optional
        List of async callbacks that take a `chunk_parser` and `index`
        as inputs and return a boolean determining whether the text
        chunk was parsed successfully or not. By default, ``None``,
        which does not use any callbacks.
    min_chunks_to_process : int, optional
        Minimum number of chunks to process before aborting due to text
        not being legal. By default, ``3``.
    """
    passed_heuristic_mem = []
    callbacks = callbacks or []
    outer_task_name = asyncio.current_task().get_name()

    for ind, text in enumerate(chunk_parser.text_chunks):
        passed_heuristic_mem.append(heuristic.check(text))
        if ind < min_chunks_to_process:
            is_legal = await legal_text_validator.check_chunk(
                chunk_parser, ind
            )
            if not is_legal:  # don't bother checking this chunk
                continue

        # don't bother checking this document
        elif not legal_text_validator.is_legal_text:
            return

        # hasn't passed heuristic, so don't pass it to callbacks
        elif not any(passed_heuristic_mem[-chunk_parser.num_to_recall :]):
            continue

        logger.debug("Processing text at ind %d", ind)
        logger.debug("Text:\n%s", text)

        if not callbacks:
            continue

        cb_futures = [
            asyncio.create_task(cb(chunk_parser, ind), name=outer_task_name)
            for cb in callbacks
        ]
        cb_results = await asyncio.gather(*cb_futures)

        # mask this chunk if we got a good result - this avoids forcing
        # the following chunk to be checked (it will only be checked if
        # it itself passes the heuristic)
        passed_heuristic_mem[-1] = not any(cb_results)



"""Ordinances usage tracking utilities"""

import time
import logging
from collections import UserDict, deque
from functools import total_ordering


logger = logging.getLogger(__name__)


@total_ordering
class TimedEntry:
    """An entry that performs comparisons based on time added, not value

    Examples
    --------
    >>> a = TimedEntry(100)
    >>> a > 1000
    True
    """

    def __init__(self, value):
        """

        Parameters
        ----------
        value : obj
            Some value to store as an entry.
        """
        self.value = value
        self._time = time.monotonic()

    def __eq__(self, other):
        return self._time == other

    def __lt__(self, other):
        return self._time < other

    def __hash__(self):
        return hash((self.value, self._time))


class TimeBoundedUsageTracker:
    """Track usage of a resource over time

    This class wraps a double-ended queue, and any inputs older than
    a certain time are dropped. Those values are also subtracted from
    the running total.

    References
    ----------
    https://stackoverflow.com/questions/51485656/efficient-time-bound-queue-in-python
    """

    def __init__(self, max_seconds=70):
        """

        Parameters
        ----------
        max_seconds : int, optional
            Maximum age in seconds of an element before it is dropped
            from consideration. By default, ``65``.
        """
        self.max_seconds = max_seconds
        self._total = 0
        self._q = deque()

    @property
    def total(self):
        """float: Total of all entries younger than `max_seconds`"""
        self._discard_old_values()
        return self._total

    def add(self, value):
        """Add a value to track

        Parameters
        ----------
        value : int or float
            A new value to add to the queue. It's total will be added to
            the running total, and it will live for `max_seconds` before
            being discarded.
        """
        self._q.append(TimedEntry(value))
        self._total += value

    def _discard_old_values(self):
        """Discard 'old' values from the queue"""
        cutoff_time = time.monotonic() - self.max_seconds
        try:
            while self._q[0] < cutoff_time:
                self._total -= self._q.popleft().value
        except IndexError:
            pass


class UsageTracker(UserDict):
    """Rate or AIP usage tracker"""

    UNKNOWN_MODEL_LABEL = "unknown_model"
    """Label used in the usage dictionary for unknown models"""

    def __init__(self, label, response_parser):
        """

        Parameters
        ----------
        label : str
            Top-level label to use when adding this usage information to
            another dictionary.
        response_parser : callable
            A callable that takes the current usage info (in dictionary
            format) and an LLm response as inputs, updates the usage
            dictionary with usage info based on the response, and
            returns the updated dictionary. See, for example,
            :func:`compass.services.openai.usage_from_response`.
        """
        super().__init__()
        self.label = label
        self.response_parser = response_parser

    def add_to(self, other):
        """Add the contents of this usage information to another dict

        The contents of this dictionary are stored under the `label`
        key that this object was initialized with.

        Parameters
        ----------
        other : dict
            A dictionary to add the contents of this one to.
        """
        other.update({self.label: {**self, "tracker_totals": self.totals}})

    @property
    def totals(self):
        """Compute total usage across all sub-labels

        Returns
        -------
        dict
            Dictionary containing usage information totaled across all
            sub-labels.
        """
        totals = {}
        for model, model_usage in self.items():
            total_model_usage = totals[model] = {}
            for report in model_usage.values():
                try:
                    sub_label_report = report.items()
                except AttributeError:
                    continue

                for tracked_value, count in sub_label_report:
                    total_model_usage[tracked_value] = (
                        total_model_usage.get(tracked_value, 0) + count
                    )
        return totals

    def update_from_model(
        self, model=None, response=None, sub_label="default"
    ):
        """Update usage from a model response

        Parameters
        ----------
        model : str, optional
            Name of model that usage is being recorded for. If ``None``
            or empty string, the usage will be placed under the
            :obj:`UsageTracker.UNKNOWN_MODEL_LABEL` label.
        response : object, optional
            Model call response, which either contains usage information
            or can be used to infer/compute usage. If ``None``, no
            update is made. By default, ``None``.
        sub_label : str, optional
            Optional label to categorize usage under. This can be used
            to track usage related to certain categories.
            By default, ``"default"``.
        """
        if response is None:
            return

        model_usage = self.setdefault(model or self.UNKNOWN_MODEL_LABEL, {})
        model_usage[sub_label] = self.response_parser(
            model_usage.get(sub_label, {}), response
        )



"""Custom Exceptions and Errors for COMPASS"""

import logging


logger = logging.getLogger("compass")


class COMPASSError(Exception):
    """Generic COMPASS Error"""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger"""
        super().__init__(*args, **kwargs)
        if args:
            logger.error(str(args[0]), stacklevel=2)


class COMPASSNotInitializedError(COMPASSError):
    """COMPASS not initialized error"""


class COMPASSValueError(COMPASSError, ValueError):
    """COMPASS ValueError"""



"""COMPASS CLI progress bars"""

import logging
from datetime import timedelta
from contextlib import contextmanager

from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    ProgressColumn,
)
from rich.console import Group
from rich.text import Text

from compass.exceptions import COMPASSValueError, COMPASSNotInitializedError


logger = logging.getLogger(__name__)


class _TimeElapsedColumn(ProgressColumn):
    """Renders time elapsed"""

    def render(self, task):  # noqa: PLR6301
        """Show time elapsed"""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return Text("[-:--:--]", style="white")
        delta = timedelta(seconds=max(0, int(elapsed)))
        return Text(f"[{delta}]", style="white")


class _MofNCompleteColumn(ProgressColumn):
    """Renders completed count/total, e.g. '   10/1000'"""

    def render(self, task):  # noqa: PLR6301
        """Show completed/total"""
        completed = int(task.completed)
        total = int(task.total) if task.total is not None else "?"
        total_width = len(str(total))
        return Text(f"   {completed:{total_width}d}/{total}", style="white")


class _TotalCostColumn(ProgressColumn):
    """Renders total cost '($1.23)'"""

    def render(self, task):  # noqa: PLR6301
        """Show completed/total"""
        total_cost = task.fields.get("total_cost", 0)
        if not total_cost:
            return Text("")
        return Text.assemble("(", (f"${total_cost:.2f}", "#71906e"), ")")


class _COMPASSProgressBars:
    """COMPASS progress bar configurations"""

    def __init__(self, console=None):
        """

        Parameters
        ----------
        console : rich.Console, optional
            Optional Console instance. Default is an internal Console
            instance writing to stdout. By default, ``None``.
        """
        self.console = console
        self._main = Progress(
            SpinnerColumn(style="dim"),
            TextColumn("{task.description}"),
            _TimeElapsedColumn(),
            BarColumn(
                complete_style="progress.elapsed",
                finished_style="progress.spinner",
            ),
            _MofNCompleteColumn(),
            _TotalCostColumn(),
            console=console,
        )
        self._group = Group(self._main)
        self._main_task = None
        self._total_cost = 0
        self._jd_pbs = {}
        self._jd_tasks = {}

    @property
    def group(self):
        """rich.console.Group: Group of renderable progress bars."""
        return self._group

    def create_main_task(self, num_jurisdictions):
        """Set up main task to track number of jurisdictions processed

        Parameters
        ----------
        num_jurisdictions : int
            Number of jurisdictions that are being processed.

        Raises
        ------
        COMPASSValueError
            If the main task has already been set up.
        """
        if self._main_task is not None:
            msg = "Main task has already been set!"
            raise COMPASSValueError(msg)

        logger.trace(
            "Starting main progress bar with %d jurisdiction(s)",
            num_jurisdictions,
        )
        if num_jurisdictions == 1:
            text = "[bold cyan]Searching 1 Jurisdiction"
        else:
            text = f"[bold cyan]Searching {num_jurisdictions:,} Jurisdictions"

        self._main_task = self._main.add_task(
            f"{text:<40}", total=num_jurisdictions
        )

    def progress_main_task(self):
        """Advance the main task one step

        In other words, mark one jurisdiction as complete.

        Raises
        ------
        COMPASSNotInitializedError
            If the main task has not been set up (i.e.
            `create_main_task` has not been called).
        """
        if self._main_task is None:
            msg = (
                "Main task has not been created! Call the "
                "`create_main_task` method first"
            )
            raise COMPASSNotInitializedError(msg)

        self._main.update(self._main_task, advance=1)

    def update_total_cost(self, cost, replace=False):
        """Update the total cost of the run

        Parameters
        ----------
        cost : int or float
            Cost value used for update.
        replace : bool, optional
            If ``True``, the `cost` input will completely replace the
            total cost, but only if the `cost` value is equal to or
            larger than the existing total cost (i.e. we never want the
            cost to decrease). If ``False``, the `cost` input is just
            added to the running total. By default, ``False``.
        """
        if replace:
            if cost + 0.01 >= self._total_cost:
                self._total_cost = cost
        else:
            self._total_cost += cost

        if self._main_task is not None:
            self._main.update(self._main_task, total_cost=self._total_cost)

    @contextmanager
    def jurisdiction_prog_bar(self, location, progress_main=True):
        """Set a progress bar for the processing of one jurisdiction

        Parameters
        ----------
        location : str
            Name of jurisdiction being processed.
        progress_main : bool, default=True
            Option to progress the main task when exiting this context
            manager.

        Yields
        ------
        rich.progress.Progress
            `rich` progress bar initialized for this jurisdiction.

        Raises
        ------
        COMPASSValueError
            If a progress bar already exists for this location.
        """
        if location in self._jd_pbs:
            msg = f"Progress bar already exists for {location}"
            raise COMPASSValueError(msg)

        pb = Progress(
            TextColumn("    "),
            SpinnerColumn(style="dim"),
            TextColumn(f"[progress.percentage]{location:<30}"),
            _TimeElapsedColumn(),
            TextColumn("[bar.back]{task.description}"),
            console=self.console,
        )
        self._jd_pbs[location] = pb
        self._group.renderables.append(pb)
        self._jd_tasks[location] = pb.add_task("")

        try:
            yield pb
        finally:
            self._remove_jurisdiction_prog_bar(location)
            if progress_main:
                self.progress_main_task()

    def _remove_jurisdiction_prog_bar(self, location):
        """Remove jurisdiction prog bar and associated task (if any)"""
        pb = self._jd_pbs.pop(location)
        if task_id := self._jd_tasks.get(location):
            pb.remove_task(task_id)

        self._group.renderables.remove(pb)

    def update_jurisdiction_task(self, location, *args, **kwargs):
        """Update the task corresponding to the jurisdiction

        Parameters
        ----------
        location : str
            Name of jurisdiction being processed.
        *args, **kwargs
            Parameters to pass to the `task.update` function in the
            `rich` python library.
        """
        task_id = self._jd_tasks[location]
        self._jd_pbs[location].update(task_id, *args, **kwargs)

    @contextmanager
    def jurisdiction_sub_prog(self, location):
        """Start a sub-progress update area for location

        This type of sub-progress does not have a bar, so it's useful
        for tasks with an unknown length/duration.

        Parameters
        ----------
        location : str
            Name of jurisdiction being processed.

        Yields
        ------
        rich.progress.Progress
            `rich` sub-progress initialized for this jurisdiction.
        """
        pb = Progress(
            TextColumn("        "),
            SpinnerColumn(style="dim"),
            TextColumn("{task.description}"),
            _TimeElapsedColumn(),
            console=self.console,
        )

        jd_pb = self._jd_pbs.get(location)
        if jd_pb:
            insert_index = self._group.renderables.index(jd_pb) + 1
        else:
            insert_index = len(self._group.renderables)

        self._group.renderables.insert(insert_index, pb)

        try:
            yield pb
        finally:
            self._group.renderables.remove(pb)

    @contextmanager
    def jurisdiction_sub_prog_bar(self, location):
        """Start a sub-progress bar for location

        Parameters
        ----------
        location : str
            Name of jurisdiction being processed.

        Yields
        ------
        rich.progress.Progress
            `rich` sub-progress bar initialized for this jurisdiction.
        """
        pb = Progress(
            TextColumn("        "),
            SpinnerColumn(style="dim"),
            TextColumn("{task.description}"),
            _TimeElapsedColumn(),
            BarColumn(
                bar_width=30,
                complete_style="progress.elapsed",
                finished_style="progress.spinner",
            ),
            _MofNCompleteColumn(),
            TextColumn("[bar.back]{task.fields[just_parsed]}"),
            console=self.console,
        )

        jd_pb = self._jd_pbs.get(location)
        if jd_pb:
            insert_index = self._group.renderables.index(jd_pb) + 1
        else:
            insert_index = len(self._group.renderables)

        self._group.renderables.insert(insert_index, pb)

        try:
            yield pb
        finally:
            self._group.renderables.remove(pb)


COMPASS_PB = _COMPASSProgressBars()
"""Compass progress bars instance (singleton)"""



"""Ordinance document content Validation logic

These are primarily used to validate that a legal document applies to a
particular technology (e.g. Solar Energy Conversion System).
"""

import logging

from compass.extraction.common import BaseTextExtractor
from compass.validation.content import Heuristic
from compass.utilities.enums import LLMUsageCategory
from compass.utilities.parsing import merge_overlapping_texts


logger = logging.getLogger(__name__)


_LARGE_SEF_SYNONYMS = (
    "solar panels, solar energy conversion systems (SECS), solar energy "
    "facilities (SEF), solar energy farms (SEF), solar farms (SF), "
    "utility-scale solar energy systems (USES), commercial solar energy "
    "systems (CSES), alternate energy systems (AES), or similar"
)
_SEARCH_TERMS_AND = (
    "zoning, siting, setback, system design, and operational "
    "requirements/restrictions"
)
_SEARCH_TERMS_OR = _SEARCH_TERMS_AND.replace("and", "or")
_IGNORE_TYPES = (
    "private, residential, roof-mounted, micro, small, or medium sized"
)


class SolarHeuristic(Heuristic):
    """Perform a heuristic check for mention of solar farms in text"""

    NOT_TECH_WORDS = ["solaris"]
    GOOD_TECH_KEYWORDS = ["solar", "setback"]
    GOOD_TECH_ACRONYMS = ["secs", "sef", "ses", "cses"]
    GOOD_TECH_PHRASES = [
        "commercial solar energy system",
        "solar energy conversion",
        "solar energy system",
        "solar panel",
        "solar farm",
        "solar energy farm",
        "utility solar energy system",
    ]


class SolarOrdinanceTextCollector:
    """Check text chunks for ordinances and collect them if they do"""

    CONTAINS_ORD_PROMPT = (
        "You extract structured data from text. Return your answer in JSON "
        "format (not markdown). Your JSON file must include exactly two "
        "keys. The first key is 'solar_reqs', which is a string that "
        f"summarizes all {_SEARCH_TERMS_AND} (if given) "
        "in the text for solar energy systems. "
        "Note that solar energy bans are an important restriction to track. "
        "The last key is '{key}', which is a boolean that is set to True if "
        f"the text excerpt describes {_SEARCH_TERMS_OR} for "
        "a solar energy system and False otherwise."
    )

    IS_UTILITY_SCALE_PROMPT = (
        "You are a legal scholar that reads ordinance text and determines "
        f"whether it applies to {_SEARCH_TERMS_OR} for "
        "large solar energy systems. Large solar energy systems (SES) may "
        f"also be referred to as {_LARGE_SEF_SYNONYMS}. "
        "Your client is a commercial solar developer that does not "
        f"care about ordinances related to {_IGNORE_TYPES} solar energy "
        "systems. Ignore any text related to such systems. "
        "Return your answer in JSON format (not markdown). Your JSON file "
        "must include exactly two keys. The first key is 'summary' which "
        "contains a string that summarizes the types of solar energy systems "
        "the text applies to (if any). The second key is '{key}', which is a "
        "boolean that is set to True if any part of the text excerpt mentions "
        f"{_SEARCH_TERMS_OR} for the large solar energy conversion "
        "systems that the client is interested in and False otherwise."
    )

    def __init__(self):
        self._ordinance_chunks = {}

    async def check_chunk(self, chunk_parser, ind):
        """Check a chunk at a given ind to see if it contains ordinance

        Parameters
        ----------
        chunk_parser : ParseChunksWithMemory
            Instance of `ParseChunksWithMemory` that contains a
            `parse_from_ind` method.
        ind : int
            Index of the chunk to check.

        Returns
        -------
        bool
            Boolean flag indicating whether or not the text in the chunk
            contains large solar energy farm ordinance text.
        """
        contains_ord_info = await chunk_parser.parse_from_ind(
            ind, self.CONTAINS_ORD_PROMPT, key="contains_ord_info"
        )
        if not contains_ord_info:
            logger.debug("Text at ind %d does not contain ordinance info", ind)
            return False

        logger.debug("Text at ind %d does contain ordinance info", ind)

        is_utility_scale = await chunk_parser.parse_from_ind(
            ind, self.IS_UTILITY_SCALE_PROMPT, key="x"
        )
        if not is_utility_scale:
            logger.debug("Text at ind %d is not for utility-scale SEF", ind)
            return False

        logger.debug("Text at ind %d is for utility-scale SEF", ind)

        _store_chunk(chunk_parser, ind, self._ordinance_chunks)
        logger.debug("Added text at ind %d to ordinances", ind)

        return True

    @property
    def contains_ord_info(self):
        """bool: Flag indicating whether text contains ordinance info"""
        return bool(self._ordinance_chunks)

    @property
    def ordinance_text(self):
        """str: Combined ordinance text from the individual chunks"""
        logger.debug(
            "Grabbing %d chunk(s) from original text at these indices: %s",
            len(self._ordinance_chunks),
            list(self._ordinance_chunks),
        )

        text = [
            self._ordinance_chunks[ind]
            for ind in sorted(self._ordinance_chunks)
        ]
        return merge_overlapping_texts(text)


class SolarPermittedUseDistrictsTextCollector:
    """Check text chunks for permitted solar districts; collect them"""

    DISTRICT_PROMPT = (
        "You are a legal scholar that reads ordinance text and determines "
        "whether the text explicitly details the districts where large "
        "solar energy farms are a permitted use. Large solar energy systems "
        f"(SES) may also be referred to as {_LARGE_SEF_SYNONYMS}. "
        "Do not make any inferences; only answer based on information that "
        "is explicitly outlined in the text. "
        "Note that relevant information may sometimes be found in tables. "
        "Return your answer in JSON format (not markdown). Your JSON file "
        "must include exactly two keys. The first key is 'districts' which "
        "contains a string that lists all of the district names for which "
        "the text explicitly permits large solar energy farms (if any). "
        "The last key is "
        "'{key}', which is a boolean that is set to True if any part of the "
        "text excerpt mentions districts where large solar energy farms"
        "are a permitted use and False otherwise."
    )

    def __init__(self):
        self._district_chunks = {}

    async def check_chunk(self, chunk_parser, ind):
        """Check a chunk to see if it contains permitted uses

        Parameters
        ----------
        chunk_parser : ParseChunksWithMemory
            Instance of `ParseChunksWithMemory` that contains a
            `parse_from_ind` method.
        ind : int
            Index of the chunk to check.

        Returns
        -------
        bool
            Boolean flag indicating whether or not the text in the chunk
            contains large solar energy farm permitted use text.
        """

        key = "contains_district_info"
        content = await chunk_parser.slc.call(
            sys_msg=self.DISTRICT_PROMPT.format(key=key),
            content=chunk_parser.text_chunks[ind],
            usage_sub_label=(
                LLMUsageCategory.DOCUMENT_PERMITTED_USE_CONTENT_VALIDATION
            ),
        )
        logger.debug("LLM response: %s", str(content))  # TODO: trace
        contains_district_info = content.get(key, False)

        if contains_district_info:
            _store_chunk(chunk_parser, ind, self._district_chunks)
            logger.debug("Text at ind %d contains district info", ind)
            return True

        logger.debug("Text at ind %d does not contain district info", ind)
        return False

    @property
    def contains_district_info(self):
        """bool: Flag indicating whether text contains district info"""
        return bool(self._district_chunks)

    @property
    def permitted_use_district_text(self):
        """str: Combined permitted use districts text from the chunks"""
        logger.debug(
            "Grabbing %d chunk(s) from original text at these indices: %s",
            len(self._district_chunks),
            list(self._district_chunks),
        )

        text = [
            self._district_chunks[ind] for ind in sorted(self._district_chunks)
        ]
        return merge_overlapping_texts(text)


class SolarOrdinanceTextExtractor(BaseTextExtractor):
    """Extract succinct ordinance text from input

    Purpose:
        Extract relevant ordinance text from document.
    Responsibilities:
        1. Extract portions from chunked document text relevant to
           particular ordinance type (e.g. solar zoning for
           utility-scale systems).
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries.
    """

    SOLAR_ENERGY_SYSTEM_FILTER_PROMPT = (
        "# CONTEXT #\n"
        "We want to reduce the provided excerpt to only contain information "
        "about solar energy systems. The extracted text will be used for "
        "structured data extraction, so it must be both **comprehensive** "
        "(retaining all relevant details) and **focused** (excluding "
        "unrelated content). Ensure that all retained information is "
        "**directly applicable** to solar energy systems while preserving "
        "full context and accuracy.\n"
        "\n# OBJECTIVE #\n"
        "Extract all text **pertaining to solar energy systems** from the "
        "provided excerpt.\n"
        "\n# RESPONSE #\n"
        "Follow these guidelines carefully:\n"
        "\n1. ## Scope of Extraction ##:\n"
        "- Include **all** text that pertains to** solar energy systems**, "
        "even if they are referred to by different names such as:\n"
        f"\t{_LARGE_SEF_SYNONYMS.capitalize()}.\n"
        "- Explicitly include any text related to **bans or prohibitions** "
        "on solar energy systems.\n"
        "\n2. ## Exclusions ##:\n"
        "- Do **not** include text that does not pertain to solar energy "
        "systems.\n"
        "- Do **not** include text that explicitly **only** applies to "
        f"{_IGNORE_TYPES} solar energy systems.\n"
        "\n3. ## Formatting & Structure ##:\n"
        "- **Preserve all section headers and numbering** for clarity and "
        "reference.\n"
        "- **Maintain the original wording, formatting, and structure** to "
        "ensure accuracy.\n"
        "\n4. ## Output Handling ##:\n"
        "- If **no relevant text** is found, return the response: "
        '"No relevant text."'
    )

    async def extract_solar_energy_system_section(self, text_chunks):
        """Extract ordinance text from input text chunks for SEF

        Parameters
        ----------
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks.

        Returns
        -------
        str
            Ordinance text extracted from text chunks.
        """
        return await self._process(
            text_chunks=text_chunks,
            instructions=self.SOLAR_ENERGY_SYSTEM_FILTER_PROMPT,
            is_valid_chunk=_valid_chunk,
        )

    @property
    def parsers(self):
        """Iterable of parsers provided by this extractor

        Yields
        ------
        name : str
            Name describing the type of text output by the parser.
        parser
            Parser that takes a `text_chunks` input and outputs parsed
            text.
        """
        yield (
            "cleaned_ordinance_text",
            self.extract_solar_energy_system_section,
        )


class SolarPermittedUseDistrictsTextExtractor(BaseTextExtractor):
    """Extract succinct ordinance text from input

    Purpose:
        Extract relevant ordinance text from document.
    Responsibilities:
        1. Extract portions from chunked document text relevant to
           particular ordinance type (e.g. solar zoning for
           utility-scale systems).
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries.
    """

    _USAGE_LABEL = LLMUsageCategory.DOCUMENT_PERMITTED_USE_DISTRICTS_SUMMARY

    PERMITTED_USES_FILTER_PROMPT = (
        "# CONTEXT #\n"
        "We want to reduce the provided excerpt to only contain information "
        "detailing permitted use(s) for a district. The extracted text will "
        "be used for structured data extraction, so it must be both "
        "**comprehensive** (retaining all relevant details) and **focused** "
        "(excluding unrelated content). Ensure that all retained information "
        "is **directly applicable** to permitted use(s) for one or more "
        "districts while preserving full context and accuracy.\n"
        "\n# OBJECTIVE #\n"
        "Remove all text **not directly pertinent** to permitted use(s) for "
        "a district.\n"
        "\n# RESPONSE #\n"
        "Follow these guidelines carefully:\n"
        "\n1. ## Scope of Extraction ##:\n"
        "- Retain all text defining permitted use(s) for a district, "
        "including:\n"
        "\t- **Primary, Special, Accessory, and other permitted use types.**\n"
        "\t- **District names and zoning classifications.**\n"
        "- Pay extra attention to any references to **solar energy "
        "facilities** or related terms.\n"
        "- Ensure that **tables, lists, and structured elements** are "
        "preserved as they may contain relevant details.\n"
        "\n2. ## Exclusions ##:\n"
        "- Do **not** include unrelated regulations, procedural details, "
        "or non-use-based restrictions.\n"
        "\n3. ## Formatting & Structure ##:\n"
        "- **Preserve all section headers and numbering** for clarity and "
        "reference.\n"
        "- **Maintain the original wording, formatting, and structure** to "
        "ensure accuracy.\n"
        "\n4. ## Output Handling ##:\n"
        "- If **no relevant text** is found, return the response: "
        '"No relevant text."'
    )

    SEF_PERMITTED_USES_FILTER_PROMPT = (
        "# CONTEXT #\n"
        "We want to reduce the provided excerpt to only contain information "
        "detailing **solar energy system** permitted use(s) for a district. "
        "The extracted text will be used for structured data extraction, so "
        "it must be both **comprehensive** (retaining all relevant details) "
        "and **focused** (excluding unrelated content). Ensure that all "
        "retained information is **directly applicable** to permitted use(s) "
        "for solar energy systems in one or more districts while "
        "preserving full context and accuracy.\n"
        "\n# OBJECTIVE #\n"
        "Remove all text **not directly pertinent** to solar energy "
        "conversion system permitted use(s) for a district.\n"
        "\n# RESPONSE #\n"
        "Follow these guidelines carefully:\n"
        "\n1. ## Scope of Extraction ##:\n"
        "- Retain all text defining permitted use(s) for a district, "
        "including:\n"
        "\t- **Primary, Special, Accessory, and other permitted use types.**\n"
        "\t- **District names and zoning classifications.**\n"
        "- Ensure that **tables, lists, and structured elements** are "
        "preserved as they may contain relevant details.\n"
        "\n2. ## Exclusions ##:\n"
        "- Do not include text that does not pertain at all to solar "
        "energy systems.\n"
        "\n3. ## Formatting & Structure ##:\n"
        "- **Preserve all section headers and numbering** for clarity and "
        "reference.\n"
        "- **Maintain the original wording, formatting, and structure** to "
        "ensure accuracy.\n"
        "\n4. ## Output Handling ##:\n"
        "- If **no relevant text** is found, return the response: "
        '"No relevant text."'
    )

    async def extract_permitted_uses(self, text_chunks):
        """Extract permitted uses text from input text chunks

        Parameters
        ----------
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks.

        Returns
        -------
        str
            Ordinance text extracted from text chunks.
        """
        return await self._process(
            text_chunks=text_chunks,
            instructions=self.PERMITTED_USES_FILTER_PROMPT,
            is_valid_chunk=_valid_chunk,
        )

    async def extract_sef_permitted_uses(self, text_chunks):
        """Extract permitted uses text for large SEF from input text

        Parameters
        ----------
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks.

        Returns
        -------
        str
            Ordinance text extracted from text chunks.
        """
        return await self._process(
            text_chunks=text_chunks,
            instructions=self.SEF_PERMITTED_USES_FILTER_PROMPT,
            is_valid_chunk=_valid_chunk,
        )

    @property
    def parsers(self):
        """Iterable of parsers provided by this extractor

        Yields
        ------
        name : str
            Name describing the type of text output by the parser.
        parser
            Parser that takes a `text_chunks` input and outputs parsed
            text.
        """
        yield "permitted_use_only_text", self.extract_permitted_uses
        yield "districts_text", self.extract_sef_permitted_uses


def _valid_chunk(chunk):
    """True if chunk has content"""
    return chunk and "no relevant text" not in chunk.lower()


def _store_chunk(parser, chunk_ind, store):
    """Store chunk and its neighbors if it is not already stored"""
    for offset in range(1 - parser.num_to_recall, 2):
        ind_to_grab = chunk_ind + offset
        if ind_to_grab < 0 or ind_to_grab >= len(parser.text_chunks):
            continue

        store.setdefault(ind_to_grab, parser.text_chunks[ind_to_grab])



"""Wind ordinance structured parsing class"""

import asyncio
import logging
from copy import deepcopy
from itertools import chain
from warnings import warn

import pandas as pd

from compass.llm.calling import BaseLLMCaller, ChatLLMCaller
from compass.extraction.features import SetbackFeatures
from compass.extraction.common import (
    EXTRACT_ORIGINAL_TEXT_PROMPT,
    run_async_tree,
    run_async_tree_with_bm,
    found_ord,
    empty_output,
    setup_async_decision_tree,
    setup_base_setback_graph,
    setup_participating_owner,
    setup_graph_extra_restriction,
    setup_graph_permitted_use_districts,
)
from compass.extraction.wind.graphs import (
    setup_graph_wes_types,
    setup_multiplier,
    setup_conditional,
)
from compass.warn import COMPASSWarning
from compass.pb import COMPASS_PB


logger = logging.getLogger(__name__)
DEFAULT_SYSTEM_MESSAGE = (
    "You are a legal scholar informing a wind energy developer about local "
    "zoning ordinances."
)
SETBACKS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on "
    "ordinances relating to setbacks from {feature} for {tech}. Ignore "
    "all text that pertains to private, micro, small, or medium sized wind "
    "energy systems."
)
RESTRICTIONS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on "
    "ordinances relating to {restriction} for {tech}. Ignore "
    "all text that pertains to private, micro, small, or medium sized wind "
    "energy systems."
)
PERMITTED_USE_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on permitted uses for "
    "{tech}. Ignore all text that pertains to private, micro, small, or "
    "medium sized wind energy systems."
)
EXTRA_NUMERICAL_RESTRICTIONS = {
    "noise": "maximum noise level allowed",
    "maximum height": "maximum turbine height allowed",
    "minimum lot size": "minimum lot, parcel, or tract size allowed",
    "shadow flicker": "maximum shadow flicker allowed",
    "tower density": "minimum turbine spacing allowed",
    "blade clearance": "minimum blade clearance allowed",
}
EXTRA_QUALITATIVE_RESTRICTIONS = {
    "color": "color or finish requirements",
    "decommissioning": "decommissioning requirements",
    "lighting": "lighting requirements",
    "moratorium": "moratoriums or bans",
    "visual impact": "visual impact assessment requirements",
}
UNIT_CLARIFICATIONS = {
    "noise": (
        "For the purposes of this extraction, assume the standard units "
        'for noise are "dBA".'
    ),
    "shadow flicker": (
        "For the purposes of this extraction, assume the standard units "
        'for shadow flicker are "hr/year".'
    ),
    "tower density": (
        "For the purposes of this extraction, assume the standard units "
        "for turbine spacing are one of the following: "
        '"tip-height-multiplier", "hub-height-multiplier", '
        '"rotor-diameter-multiplier", "feet", or "meters".'
    ),
}
ER_CLARIFICATIONS = {}


class StructuredWindParser(BaseLLMCaller):
    """Base class for parsing structured data"""

    def _init_chat_llm_caller(self, system_message):
        """Initialize a ChatLLMCaller instance for the DecisionTree"""
        return ChatLLMCaller(
            self.llm_service,
            system_message=system_message,
            usage_tracker=self.usage_tracker,
            **self.kwargs,
        )

    async def _check_wind_turbine_type(self, text):
        """Get the largest turbine size mentioned in the text"""
        logger.debug("Checking turbine_types")
        tree = setup_async_decision_tree(
            setup_graph_wes_types,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(DEFAULT_SYSTEM_MESSAGE),
        )
        decision_tree_wes_types_out = await run_async_tree(tree)

        return (
            decision_tree_wes_types_out.get("largest_wes_type")
            or "large wind energy systems"
        )


class StructuredWindOrdinanceParser(StructuredWindParser):
    """LLM ordinance document structured data scraping utility

    Purpose:
        Extract structured ordinance data from text.
    Responsibilities:
        1. Extract ordinance values into structured format by executing
           a decision-tree-based chain-of-thought prompt on the text for
           each value to be extracted.
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries and multiple
        :class:`~compass.extraction.tree.AsyncDecisionTree` instances
        to guide the extraction of individual values.
    """

    async def parse(self, text):
        """Parse text and extract structure ordinance data

        Parameters
        ----------
        text : str
            Ordinance text which may or may not contain setbacks for one
            or more features (property lines, structure, roads, etc.).
            Text can also contain other supported regulations (noise,
            shadow-flicker, etc,) which will be extracted as well.

        Returns
        -------
        pd.DataFrame
            DataFrame containing parsed-out ordinance values.
        """
        largest_wes_type = await self._check_wind_turbine_type(text)
        logger.info("Largest WES type found in text: %s", largest_wes_type)

        outer_task_name = asyncio.current_task().get_name()
        with COMPASS_PB.jurisdiction_sub_prog_bar(outer_task_name) as sub_pb:
            task_id = sub_pb.add_task(
                "Extracting ordinance values...",
                total=len(SetbackFeatures.DEFAULT_FEATURE_DESCRIPTIONS)
                + len(EXTRA_NUMERICAL_RESTRICTIONS)
                + len(EXTRA_QUALITATIVE_RESTRICTIONS),
                just_parsed="",
            )
            outputs = await self._parse_all_restrictions_with_pb(
                sub_pb, task_id, text, largest_wes_type, outer_task_name
            )
            sub_pb.remove_task(task_id)

        return pd.DataFrame(chain.from_iterable(outputs))

    async def _parse_all_restrictions_with_pb(
        self, sub_pb, task_id, text, largest_wes_type, outer_task_name
    ):
        """Parse all ordinance values"""
        feature_parsers = [
            asyncio.create_task(
                self._parse_setback_feature(
                    sub_pb, task_id, text, feature_kwargs, largest_wes_type
                ),
                name=outer_task_name,
            )
            for feature_kwargs in SetbackFeatures()
        ]
        extras_parsers = [
            asyncio.create_task(
                self._parse_extra_restriction(
                    sub_pb,
                    task_id,
                    text,
                    feature,
                    r_text,
                    largest_wes_type,
                    is_numerical=True,
                    unit_clarification=UNIT_CLARIFICATIONS.get(feature, ""),
                    feature_clarifications=ER_CLARIFICATIONS.get(feature, ""),
                ),
                name=outer_task_name,
            )
            for feature, r_text in EXTRA_NUMERICAL_RESTRICTIONS.items()
        ]
        extras_parsers += [
            asyncio.create_task(
                self._parse_extra_restriction(
                    sub_pb,
                    task_id,
                    text,
                    feature,
                    r_text,
                    largest_wes_type,
                    is_numerical=False,
                    feature_clarifications=ER_CLARIFICATIONS.get(feature, ""),
                ),
                name=outer_task_name,
            )
            for feature, r_text in EXTRA_QUALITATIVE_RESTRICTIONS.items()
        ]
        return await asyncio.gather(*(feature_parsers + extras_parsers))

    async def _parse_extra_restriction(
        self,
        sub_pb,
        task_id,
        text,
        feature,
        restriction_text,
        largest_wes_type,
        is_numerical,
        unit_clarification="",
        feature_clarifications="",
    ):
        """Parse a non-setback restriction from the text"""
        logger.debug("Parsing extra feature %r", feature)
        system_message = RESTRICTIONS_SYSTEM_MESSAGE.format(
            restriction=restriction_text, tech=largest_wes_type
        )
        tree = setup_async_decision_tree(
            setup_graph_extra_restriction,
            is_numerical=is_numerical,
            tech=largest_wes_type,
            restriction=restriction_text,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            unit_clarification=unit_clarification,
            feature_clarifications=feature_clarifications,
        )
        info = await run_async_tree(tree)
        info.update({"feature": feature, "quantitative": is_numerical})
        if is_numerical:
            info = _sanitize_output(info)
        sub_pb.update(task_id, advance=1, just_parsed=feature)
        return [info]

    async def _parse_setback_feature(
        self, sub_pb, task_id, text, feature_kwargs, largest_wes_type
    ):
        """Parse values for a setback feature"""
        feature = feature_kwargs["feature_id"]
        feature_kwargs["tech"] = largest_wes_type
        logger.debug("Parsing feature %r", feature)

        base_messages = await self._base_messages(text, **feature_kwargs)
        if not found_ord(base_messages):
            logger.debug("Failed `found_ord` check for feature %r", feature)
            sub_pb.update(task_id, advance=1, just_parsed=feature)
            return empty_output(feature)

        if feature not in {"structures", "property line"}:
            output = {"feature": feature}
            output.update(
                await self._extract_setback_values(
                    text,
                    base_messages=base_messages,
                    **feature_kwargs,
                )
            )
            sub_pb.update(task_id, advance=1, just_parsed=feature)
            return [output]

        output = await self._extract_setback_values_for_p_or_np(
            text, base_messages, **feature_kwargs
        )
        sub_pb.update(task_id, advance=1, just_parsed=feature)
        return output

    async def _base_messages(self, text, **feature_kwargs):
        """Get base messages for setback feature parsing"""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(
            feature=feature_kwargs["feature"],
            tech=feature_kwargs["tech"],
        )
        tree = setup_async_decision_tree(
            setup_base_setback_graph,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **feature_kwargs,
        )
        await run_async_tree(tree, response_as_json=False)
        return deepcopy(tree.chat_llm_caller.messages)

    async def _extract_setback_values_for_p_or_np(
        self, text, base_messages, **feature_kwargs
    ):
        """Extract setback values for participating ordinances"""
        logger.debug("Checking participating vs non-participating")
        decision_tree_participating_out = await self._run_setback_graph(
            setup_participating_owner,
            text,
            base_messages=deepcopy(base_messages),
            **feature_kwargs,
        )
        outer_task_name = asyncio.current_task().get_name()
        p_or_np_parsers = [
            asyncio.create_task(
                self._parse_p_or_np_text(
                    key, sub_text, base_messages, **feature_kwargs
                ),
                name=outer_task_name,
            )
            for key, sub_text in decision_tree_participating_out.items()
        ]
        return await asyncio.gather(*p_or_np_parsers)

    async def _parse_p_or_np_text(
        self, key, sub_text, base_messages, **feature_kwargs
    ):
        """Parse participating sub-text for ord values"""
        feature = feature_kwargs["feature_id"]
        out_feat_name = f"{feature} ({key})"
        output = {"feature": out_feat_name}
        if not sub_text:
            return output

        feature = feature_kwargs["feature"]
        feature = f"{key} {feature}"
        feature_kwargs["feature"] = feature

        base_messages = deepcopy(base_messages)
        base_messages[-2]["content"] = EXTRACT_ORIGINAL_TEXT_PROMPT.format(
            feature=feature, tech=feature_kwargs["tech"]
        )
        base_messages[-1]["content"] = sub_text

        values = await self._extract_setback_values(
            sub_text,
            base_messages=base_messages,
            **feature_kwargs,
        )
        output.update(values)
        return output

    async def _extract_setback_values(self, text, **kwargs):
        """Extract setback values for a given feature from input text"""
        decision_tree_out = await self._run_setback_graph(
            setup_multiplier, text, **kwargs
        )
        decision_tree_out = _update_output_keys(decision_tree_out)
        decision_tree_out = _sanitize_output(decision_tree_out)

        if decision_tree_out.get("value") is None:
            return decision_tree_out

        decision_tree_conditional_out = await self._run_setback_graph(
            setup_conditional, text, **kwargs
        )
        decision_tree_out.update(decision_tree_conditional_out)
        return decision_tree_out

    async def _run_setback_graph(
        self,
        graphs_setup_func,
        text,
        feature,
        tech,
        base_messages=None,
        **kwargs,
    ):
        """Generic function to run async tree"""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(
            feature=feature, tech=tech
        )
        tree = setup_async_decision_tree(
            graphs_setup_func,
            feature=feature,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **kwargs,
        )
        if base_messages:
            return await run_async_tree_with_bm(tree, base_messages)
        return await run_async_tree(tree)


class StructuredWindPermittedUseDistrictsParser(StructuredWindParser):
    """LLM permitted use districts  scraping utility

    Purpose:
        Extract structured ordinance data from text.
    Responsibilities:
        1. Extract ordinance values into structured format by executing
           a decision-tree-based chain-of-thought prompt on the text for
           each value to be extracted.
    Key Relationships:
        Uses a :class:`~compass.llm.calling.StructuredLLMCaller` for
        LLM queries and multiple
        :class:`~compass.extraction.tree.AsyncDecisionTree` instances
        to guide the extraction of individual values.
    """

    _LARGE_WES_CLARIFICATION = (
        "Large wind energy systems (WES) may also be referred to as wind "
        "turbines, wind energy conversion systems (WECS), wind energy "
        "facilities (WEF), wind energy turbines (WET), large wind energy "
        "turbines (LWET), utility-scale wind energy turbines (UWET), "
        "commercial wind energy conversion systems (CWECS), alternate "
        "energy systems (AES), or similar. "
    )
    _USE_TYPES = [
        {
            "feature_id": "primary use districts",
            "use_type": "primary use or similar (e.g., without special "
            "conditions or approval)",
        },
        {
            "feature_id": "special use districts",
            "use_type": "special use or similar (e.g., requires approval "
            "by the zoning appeals board or meeting certain conditions like "
            "completing a permitting process)",
        },
        {
            "feature_id": "accessory use districts",
            "use_type": "accessory use or similar (e.g., when integrated "
            "with an existing structure or secondary to another use)",
        },
    ]

    async def parse(self, text):
        """Parse text and extract permitted use districts data

        Parameters
        ----------
        text : str
            Permitted use districts text which may or may not contain
            information about allowed uses in one or more districts.

        Returns
        -------
        pd.DataFrame
            DataFrame containing parsed-out allowed-use district names.
        """
        largest_wes_type = await self._check_wind_turbine_type(text)
        logger.info("Largest WES type found in text: %s", largest_wes_type)

        outer_task_name = asyncio.current_task().get_name()
        with COMPASS_PB.jurisdiction_sub_prog_bar(outer_task_name) as sub_pb:
            task_id = sub_pb.add_task(
                "Extracting permitted uses...",
                total=len(self._USE_TYPES),
                just_parsed="",
            )
            feature_parsers = [
                asyncio.create_task(
                    self._parse_permitted_use_districts(
                        sub_pb,
                        task_id,
                        text,
                        largest_wes_type,
                        **use_type_kwargs,
                    ),
                    name=outer_task_name,
                )
                for use_type_kwargs in self._USE_TYPES
            ]
            outputs = await asyncio.gather(*(feature_parsers))

            sub_pb.remove_task(task_id)

        return pd.DataFrame(chain.from_iterable(outputs))

    async def _parse_permitted_use_districts(
        self, sub_pb, task_id, text, largest_wes_type, feature_id, use_type
    ):
        """Parse a non-setback restriction from the text"""
        logger.debug("Parsing use type: %r", feature_id)
        system_message = PERMITTED_USE_SYSTEM_MESSAGE.format(
            tech=largest_wes_type
        )
        tree = setup_async_decision_tree(
            setup_graph_permitted_use_districts,
            tech=largest_wes_type,
            clarifications=self._LARGE_WES_CLARIFICATION,
            text=text,
            use_type=use_type,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
        )
        info = await run_async_tree(tree)
        sub_pb.update(task_id, advance=1, just_parsed=feature_id)
        info.update({"feature": feature_id, "quantitative": True})
        return [info]


def _update_output_keys(output):
    """Standardize output keys

    We could standardize output keys by modifying the LLM prompts, but
    have found that it's more accurate to instruct the LLM to use
    descriptive keys (e.g. "mult_value" instead of "value" or
    "mult_type" instead of "units")
    """

    if "mult_value" not in output:
        return output

    output["value"] = output.pop("mult_value")

    if units := output.get("units"):
        msg = f"Found non-null units value for multiplier: {units}"
        warn(msg, COMPASSWarning)
    output["units"] = output.pop("mult_type", None)

    return output


def _sanitize_output(output):
    """Perform some sanitization on outputs"""
    return _remove_units_for_empty_value(output)


def _remove_units_for_empty_value(output):
    """Remove units if no value found"""
    units = output.get("units")
    if not units:
        return output

    value = output.get("value")
    if value:
        return output

    # at this point, we have units but no value, so remove units
    output["units"] = None
    return output
